// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_FERMI_H_
#define _SX_FERMI_H_

#include <SxTypes.h>
#include <SxKPoints.h>
#include <SxBinIO.h>
#include <SxDFT.h>

/** brief Fermi distribution function

  \b SxFermi = S/PHI/nX Fermi distribution

  In order to treat systems ate finite temperatures a Fermi like 
  function has to be
  
  \ingroup group_dft
  \author  Sixten Boeck
  */
class SX_EXPORT_DFT SxFermi
{
   public:
      
      bool    printHartree;
      Focc    focc;
      Eps     eps;
      mutable EpsSort epsSortList;
      Real8   nElectrons;
      int     nSpin;
      Real8   fFull;
      Real8   eFermi;
      Real8   spinMoment;
      Real8   eqEntropy;
      Real8   noneqEntropy;
      bool    keepSpinMoment;

      /// Whether to use occupation numbers or not
      enum FoccHandling { NoFocc, UseFocc};

      const SxKPoints *kpPtr;

      SxList<Real8> eSpinFermi; //:iSpin
      
      SxFermi ();
      SxFermi (Real8 nElectrons, int nStates, int nSpin, const SxKPoints &);
      SxFermi (int nStates, int nSpin, int nk);
      SxFermi (const SxBinIO &io, bool keepNStates = false) { 
         kpPtr = NULL; 
         nSpin = -1;
         nElectrons = -1; 
         read (io, keepNStates);};
      ~SxFermi ();

      void setOccupencies (const Focc &initialFocc);

   protected:
      /** \brief Get index-list from symbol table.
          Parses range[from, to] and values=[...]
          Negative indices are interpreted as from-end-counting.

          \note Auxiliary routine for readOccupations
        */
      static SxList<int> getIdxList(const SxSymbolTable *table,
                                     int nMax,
                                     const SxString &name);
   public:
      /** \brief Read occupation numbers from symbol table
          \note: the size must be initialized
        */
      void readOccupations (const SxSymbolTable *);
      void setSpinMoment (Real8 spinMoment, bool keepFixed=false);
      void fermiDistribution (Real8 ekt, Real8 mixFactor=1.0);
      void printOccupation (bool final=false) const;
      /** \brief Print one k-point
           @param iSpin spin index
           @param ik    k-point index
           @param printOcc whether to print the occupation numbers, too
        */
      void printK (int iSpin, int ik, bool printOcc, bool final=false) const;

      SxBundle3<Int> valBandIdx;
      SxBundle3<Int> conBandIdx;
      SxList<SxList<int> > nValBands, nConBands;  // :ik,:iSpin

      /// Get number of states 
      int getNStates (int ik = 0) const;
      /// Get number of k-points
      int getNk      () const;
      /// Get number of spin channels
      inline int getNSpin () const { return nSpin; }

      int getNValenceBands     (int iSpin, int ik) const;
      /// Get max. number of valence bands for all k and spin
      int getNValenceBands     () const;
      /// Get max. number of conduction bands for all k and spin
      int getNConductionBands  () const;
      int getNConductionBands  (int iSpin, int ik) const;
      int getValenceBandIdx    (int i, int iSpin, int ik) const;
      int getConductionBandIdx (int i, int iSpin, int ik) const;
      /**
        \brief Is this a semiconductor?

        The criteria for a semiconductor we use here is that the number
        of occupied states is the same for all k-points. It is furthermore
        checked that the number of occupied and unoccupied states is the
        total number of states.
        It is crucial that updateValCon () has been called before this 
        routine.
        */
      bool isSemiconductor ();
      double getBandGap (int *maxValIk=NULL, int *minConIk=NULL);
      double getEBand (enum FoccHandling) const;

      /** \brief Update number of occupied and unoccupied bands
        \param fOccMin   minimum occupation to regard band as occupied
        \param fUnoccMax maximum occupation to regard band as unoccupied
        \note At finite temperature, a band may be 
              considered occupied and unoccupied at the same time.
        \note For nSpin == 2, the fUnoccMax factor is multiplied by 2!
      */
      void updateValCon (double fOccMin   =    1e-7, 
                         double fUnoccMax = 1.-1e-7);
      void checkOccupation ();
      /// Write spectrum to ascii files
      void writeSpectrum (const SxString &filebase, 
                          const SxString &fileext) const;
      /// Read spectrum from ascii file
      void readSpectrum (const SxString &file,
                         SxCell *cellPtr = NULL,
                         SxKPoints *kp = NULL,
                         int iSpin = 0);
      /** \brief Try to extract nk and nStates from a spectrum file (eps.dat)
        \param epsFile    filename
        \param nkPtr      pointer to nk
        \param nStatesPtr pointer to nStates
        */
      static void peekSpectrumFile (const SxString &epsFile,
                                    int *nkPtr, int *nStatesPtr);
                         
      void write (const SxBinIO &) const;
      void read (const SxBinIO &, bool keepNStates = false);
      Real8 getSpinMoment ();
      Real8 getEntropy ();

      Real8 fermiFunction (Real8 energy, Real8 ekt, Real8 nEl, int spin);
      /** 
        \todo Error handling if root of Fermi function can not
              be found.  */
      Real8 getFermiEnergy (Real8 ekt, Real8 nEl, int spin=0); 

      /// \todo rename: resize
      void resizeFermi (int nStates);
      /// \todo rename: resize
      void resizeFermi (const SxArray<int> &nPerK);

      Focc getFoccByWindow (const double lowEnergy, const double highEnergy, const double ekt);

      int getHOMO ();

   protected:

      void init ();
};


#endif /* _SX_FERMI_H_ */
