// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxRho.h>
#include <SxGkBasis.h>
#include <SxRadBasis.h>
#include <SxProjector.h>
#include <SxQuantumNumbers.h>
#include <SxConstants.h>
#include <SxYlm.h>
#include <SxLoopMPI.h>  // LoopMPI
#include <SxPAWSet.h>

SxRho::SxRho ()
{
   rBasisPtr = NULL;
   nElectrons = 0.;
}

SxRho::SxRho (const RhoR &rhoRIn)
{
   rhoR       = rhoRIn;
   rBasisPtr  = dynamic_cast<const SxRBasis *>(rhoRIn(0).getBasisPtr());
   SX_CHECK (rBasisPtr);
   nElectrons = getNorm ();
}
SxRho::SxRho (const SxMeshR &rhoIn)
{
   rhoR.resize (1);
   rhoR(0) = rhoIn;
   rBasisPtr  = dynamic_cast<const SxRBasis *>(rhoIn.getBasisPtr());
   SX_CHECK (rBasisPtr);
   nElectrons = getNorm ();
}

SxRho::SxRho (const SxRBasis &R, int nSpin, Real8 nElectrons_)
{
   SX_CHECK_VAR (nSpin == 1 || nSpin == 2, nSpin);
   rBasisPtr = &R;
   nElectrons = nElectrons_;
   rhoR.resize (nSpin);
   for (int iSpin=0; iSpin < nSpin; ++iSpin) 
      rhoR(iSpin).resize (R.getMeshSize());
}

SxRho::SxRho (SxBinIO &io, const SxRBasis* rBasisPtrIn)
   : rBasisPtr(rBasisPtrIn)
{
   SX_CHECK (io.mode == SxBinIO::BINARY_READ_ONLY);
   int nSpin;
   try  {
      nSpin = io.getDimension ("nMeshes");
      SX_CHECK_VAR(nSpin > 0, nSpin);

      if (rBasisPtr)  {
         rhoR.resize (nSpin);
         readRho (io);
      } else {

         // --- read meshes
         SxMatrix3<Double> cell;
         SxVector3<Int>    dim;
         rhoR = io.readMesh (&cell, &dim);

         // --- determine nElectrons
         nElectrons = 0.;
         for (int iSpin = 0; iSpin < nSpin; iSpin++)
            nElectrons += rhoR(iSpin).sum ();
         
         double dOmega = fabs(cell.determinant ()) / double(dim.product ());

         nElectrons *= dOmega;

      }
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }

}

SxRho::~SxRho ()
{
   // empty
}

void SxRho::operator= (const SxRho &in)
{
   rBasisPtr  = in.rBasisPtr;
   rhoR       = in.rhoR;
   nElectrons = in.nElectrons;
}

SxMeshR &SxRho::operator() (int iSpin)
{
   return rhoR(iSpin);
}

const SxMeshR &SxRho::operator() (int iSpin) const
{
   return rhoR(iSpin);
}

const SxBasis *SxRho::getBasisPtr () const
{
   return rhoR(0).getBasisPtr();
}

const SxRBasis *SxRho::getRBasisPtr () const
{
   return rBasisPtr;
}


RhoR &SxRho::computeRho (const Focc &focc, const SxPWSet &psiSet)
{
   SX_CHECK (rBasisPtr);
   SX_CLOCK (Timer::Rho);

   // --- basis sets
   const SxGkBasis &gk = psiSet.getGkBasis ();  // |G+k>
   const SxRBasis  &R  = *rBasisPtr;            // |R>

   int i, nStates, iSpin, ik;
   int nSpin = psiSet.getNSpin ();

   // --- initialize rhoR
   for (iSpin=0; iSpin < nSpin; iSpin++)  {
      rhoR(iSpin).resize (R.getMeshSize());
      rhoR(iSpin).set (0.);
      rhoR(iSpin).setBasis (&R);
   }
   bool blockStates =   dynamic_cast<const SxPW*>(&psiSet)
                     || dynamic_cast<const SxPAWSet*>(&psiSet);

   // --- compute density
   SxDiracVec<TPrecCoeffG> psiR;
   double f;
   SX_NEW_LOOP (psiSet);
   for (ik=0; ik < gk.nk; ik++)  {
      for (iSpin=0; iSpin < nSpin; iSpin++)  {
         SX_MPI_LEVEL("waves-k");
         if (SxLoopMPI::myWork(ik)) {  // SxParallelHierarchy
            nStates = minimum (psiSet.getNStates (ik), focc.getNStates(ik));
            if (blockStates && gk(ik).hasMixedFFT ())  {
               SX_CHECK_VARS (psiSet.getNStates (ik) <= focc.getNStates(ik),
                              psiSet.getNStates (ik), focc.getNStates(ik));

               gk(ik).addToRho (psiSet(iSpin, ik),
                                gk.weights(ik) * focc(iSpin,ik),
                                 &rhoR(iSpin));
            } else {
               for (i=0; i < nStates; i++)  {
                  if ((f = gk.weights(ik) * focc(i,iSpin,ik)) > 1e-12)  {
                     /*
                  rhoR(iSpin) += (gk.weights(ik) * focc(i,iSpin,ik))
                      * (R | psiSet(i,iSpin,ik) ).absSqr();
                      */
                     psiR = (R | gk(ik)) * (gk(ik) | psiSet(i,iSpin,ik));
                     rhoR(iSpin).plus_assign_ax (f, psiR.absSqr());

                  }
               }
            }
         }  // LoopMPI
      }
   }

   for (iSpin=0; iSpin < nSpin; iSpin++)  {
      SX_MPI_SOURCE("waves-k", TaskGroupMaster);
      SX_MPI_TARGET(TopLevel, TaskGroupAll);
      SxLoopMPI::sum (rhoR(iSpin));
      rhoR(iSpin) = R.symmetrize (rhoR(iSpin));
   }

   checkNorm (rhoR);

   return rhoR;
}


RhoR SxRho::computeRho (const Focc &/*focc*/,
                        const SxPW &/*psiSet0*/, const SxPW &/*psiSet1*/)
{
   RhoR rhoR;
   SX_EXIT;  // to be implemented
   return rhoR;
}

/*
RhoR &SxRho::atomicChargeDensity (const SxAtomicStructure &str, 
                                  const SxPseudoPot &pot,
                                  Real8 spinMoment)
{
   int nSpin = rhoR.getSize();

   SX_CHECK      (rBasisPtr);
   SX_CHECK      (rBasisPtr->gBasis);  // is |G> registered to |R>???
   SX_CHECK_VARS ((nSpin == 1 && fabs(spinMoment) < 1e-10) || (nSpin == 2),
                nSpin, spinMoment);

   // --- basis sets
   double omega = str.cell.volume;
   const SxRBasis &R = *rBasisPtr;                  // |G>
   const SxGBasis &G = *R.gBasis;                   // |R>
   SxRadBasis r (pot.pseudoRad, pot.logDr, omega);  // |r>

   SxMeshG rhoG (G.ng, 0.); rhoG.setBasis (&G);
   SxGBasis::TPsi psiAtom;

   Real8 focc, nEl;
   Real8 logDr=0.;
   SxDiracVec<TReal8> psRad, psPsi, radFun;
   radFun.setBasis (&r);

   SxDiracVec<TPrecPhase>  S; // structure factor

   int iSpecies, l, maxR, size;
   int iSpin;
   // --- we use initial magn. moment for rho construction only if 
   //     there are no species with explicitly set 
   //     spinUp and spinDown occupations
   bool useSpinMoment = true;  

   int s = SxQuantumNumbers::s;
   
   for (iSpin=0; iSpin < nSpin; iSpin++)  {     
      for (iSpecies=0; iSpecies < pot.nSpecies; iSpecies++)  {
         nEl = pot.valenceCharge(iSpecies);
         for (l=0, maxR=0; l <= pot.lMax(iSpecies); l++)  {
            size = pot.pseudoRad (iSpecies)(l).getSize();
            maxR = size > maxR ? size : maxR;
         }
         radFun.resize (maxR); radFun.set(0.);
         for (l=0; l <= pot.lMax(iSpecies); l++)  {
            logDr   = pot.logDr(iSpecies)(l);
            psRad   = pot.pseudoRad(iSpecies)(l);
            psPsi   = pot.pseudoPsi(iSpecies)(l);
            
            // --- if both up & down occ.numbers are set
            if (pot.foccAtomicRho(iSpecies).getSize() == nSpin)  { 
               focc = pot.foccAtomicRho(iSpecies)(iSpin)(l);
               useSpinMoment = false; 
            }  else  {
               // --- if spin used but only one occ. number channel is set 
               //     electrons are divided between spin channels as 50/50
               if (pot.foccAtomicRho(iSpecies).getSize() > 0 )  { 
                  focc = pot.foccAtomicRho(iSpecies)(0)(l) 
                     / (Real8)nSpin;  
                  // --- if no occupation numbers are set, use default partition
               }  else  {
                  if (pot.pseudoFocc(iSpecies)(l))
                     // --- focc = 2,6,10,... if no spin and 1,3,5,... otherwise 
                     focc = minimum (4.*l+2., nEl) / (Real8)nSpin; 
                  else 
                     focc = 0.;
               }
            }
            nEl    -= focc;
            radFun += focc * (psPsi.sqr());
            
         }
         radFun  *= SQRT_1_4PI; // = Y(0,0)
         psRad.setBasis (&r);  // TODO: ugly!!!
         SxDiracVec<Double> res = SxDiracVec<Double> (radFun.getSize());
         res.handle->auxData = radFun.handle->auxData;
         sxprintf ("--- %p/%p/%p/%p\n", psRad.getBasisPtr(),
                                   radFun.getBasisPtr(),
                                   res.getBasisPtr(),
                                   radFun.handle->auxData.basisPtr);
         cout << "AO density should be projected Dirac-like" << endl;
         psiAtom  = r.toPWBasis  (psRad,radFun, G, SxQuantumNumbers::s, logDr);
         psiAtom *= r.getYlmNormFactor (s,0) * r.rNorm; // normalization

         ////       SUM_{r} <G | r> * <r | radFun *   Y_lm(s) >
         //radFun.handle->auxData.is = iSpecies; // TODO: ugly!!!
         //radFun.handle->auxData.l  = s;        // TODO: ugly!!!
         //radFun.handle->auxData.m  = 0;        // TODO: ugly!!!

         //psiAtom  = SUM(r,  (G | r) * (r | radFun * r.Ylm(s)  ) );
         // translator T = G.structureFactors(iSpecies);

         //S = G.structureFactors(iSpecies);
         //      e^(-iGkr) * Sum_r   < G | r > < r | radFun *  Y_sm    >
         //rhoG +=   S      * SUM (r, ( G | r )*( r | radFun * r.Ylm(s) ) );
									
         rhoG += G.structureFactors(iSpecies) * psiAtom;
      }
      //   rho(R) = < R | rho(G) >
      rhoR(iSpin) = ( R | rhoG   );
      rhoG = 0.;
   }

   if (nSpin == 2)  {
      if (useSpinMoment)  {
         Real8 nEl = nElectrons / (Real8)nSpin;
         rhoR(SPIN_UP)   *= 1. + spinMoment/nEl;
         rhoR(SPIN_DOWN) *= 1. - spinMoment/nEl;
      }  
   }
   for (iSpin=0; iSpin < nSpin; iSpin++)   
      rhoR(iSpin) = R.symmetrize (rhoR(iSpin));

   checkNorm ();
   normalizeRho ();

   return rhoR;
}
*/


// --- *** UGLY QUICK&DIRTY COPY FROM REL-1-0 ***
//     *** TOBE REWRITEN IN A CLEAR WAY LATER *** {
RhoR &SxRho::atomicChargeDensity (const SxAtomicStructure &str, 
                                  const SxPseudoPot       &pot)
{
   SX_CHECK (rBasisPtr);
   
   // --- basis sets
   const SxRBasis &R = *rBasisPtr;                  // |R>
   const SxGBasis &G = R.getGBasis ();              // |G>
   SxRadBasis r (pot.rad, pot.logDr);  // |r>
   SX_CHECK (str.cell.volume > 0.);

   SxMeshG rhoG (G.ng, 0.); rhoG.setBasis (&G);
   SxGBasis::TPsi psiAtom;

   Real8 focc;
   Real8 logDr=0.;
   SxDiracVec<TReal8> psRad, psPsi, radFun;
   radFun.setBasis (&r);

   SxDiracVec<TPrecPhase>  S; // structure factor

   int iSpecies, l;
   int iSpin, nSpin = getNSpin ();

   int s = SxQuantumNumbers::s;
  
   // --- print-out
   cout << SX_SEPARATOR;
   cout << "|  Initial occupations\n";
   cout << SX_SEPARATOR;
   for (iSpecies=0; iSpecies < pot.nSpecies; iSpecies++)  {
      cout << "| " << pot.elementName(iSpecies) << ":  ";
      for (l=0; l <= pot.lMax(iSpecies); l++)  {
         switch (l)  {
            case 0 : sxprintf ("  s="); break;
            case 1 : sxprintf ("  p="); break;
            case 2 : sxprintf ("  d="); break;
            case 3 : sxprintf ("  f="); break;           
            default: sxprintf ("  %d=", l);
         }
         for (iSpin=0; iSpin < nSpin; iSpin++)  {  
            focc = pot.foccAtomicRho(iSpecies)(l)(iSpin); 
            if (iSpin == 0)  sxprintf ("%2.2fe",focc);
            else             sxprintf ("/%2.2fe",focc);
         }
      }
      sxprintf ("\n");
   }	
   cout << SX_SEPARATOR;

   for (iSpin=0; iSpin < nSpin; iSpin++)  {
      for (iSpecies=0; iSpecies < pot.nSpecies; iSpecies++)  {
         int nr = (int)pot.rad(iSpecies).getSize ();
         radFun.resize (nr); radFun.set(0.);
         logDr   = pot.logDr(iSpecies);
         psRad   = pot.rad(iSpecies);
         for (l=0; l <= pot.lMax(iSpecies); l++)  {
            psPsi   = pot.pseudoPsi(iSpecies)(l);
            focc    = pot.foccAtomicRho(iSpecies)(l)(iSpin); 
            radFun += focc * (psPsi.sqr());
         }
         radFun  *= SQRT_1_4PI; // = Y(0,0)
         psRad.setBasis (&r);  // TODO: ugly!!!
         SxDiracVec<Double> res = SxDiracVec<Double> (radFun.getSize());
         res.handle->auxData = radFun.handle->auxData;
         //cout << "AO density should be projected Dirac-like" << endl;
         psiAtom  = r.toPWBasis  (psRad,radFun, G, SxQuantumNumbers::s, logDr);
         psiAtom *= SxYlm::getYlmNormFactor (s,0) * FOUR_PI / sqrt(str.cell.volume); // normalization
         rhoG += G.structureFactors(iSpecies) * psiAtom;
      }
      rhoR(iSpin) = ( R | rhoG   );
      rhoG.set(0.);
   }

   for (iSpin=0; iSpin < nSpin; iSpin++)   
      rhoR(iSpin) = R.symmetrize (rhoR(iSpin));

   checkNorm ();
   normalizeRho ();

   return rhoR;
}
// --- *** UGLY QUICK&DIRTY COPY FROM REL-1-0 ***
//     *** TOBE REWRITEN IN A CLEAR WAY LATER *** }



//------------------------------------------------------------------------------
// Check for correct norm of charge density
//------------------------------------------------------------------------------
void SxRho::checkNorm ()
{
   checkNorm (rhoR);
}


Real8 SxRho::getNorm () const
{
   Real8 norm = 0.;
   for (int iSpin=0; iSpin < rhoR.getSize (); iSpin++)  {
      norm += rhoR(iSpin).sum() * rBasisPtr->dOmega;
   }
   return norm;
}


void SxRho::normalizeRho ()
{
   normalizeRho (&rhoR, getNorm ());
}


void SxRho::randomize ()
{
   for (int iSpin=0; iSpin < rhoR.getSize(); ++iSpin)  {
      rhoR(iSpin).randomize ();
      rhoR(iSpin).setBasis (rBasisPtr);
   }
   normalizeRho ();
}

void SxRho::checkNorm (RhoR &rhoIn) const
{
   if (nElectrons > 1e-10)  {
      Real8 norm = getNorm ();
      if (fabs (norm - nElectrons) > 0.001)  {
         cout << "\nThe norm of charge density is getting lost. " << endl;
         cout << "Norm = " << norm;
         cout << " (should be " << nElectrons << ")" << endl;
         normalizeRho (&rhoIn, norm);
      }
   }  else  {
      // sxprintf ("Skipping rho norm check.\n");
   }
}

void SxRho::normalizeRho (RhoR *rhoIn, Real8 rhoInNorm) const
{
   for (int iSpin=0; iSpin < rhoIn->getSize(); iSpin++)
      (*rhoIn)(iSpin) *= nElectrons/rhoInNorm;
}

//SxRho &SxRho::mixRhoLinearly (const RhoR &rhoIn, Real8 mixFactor)
//{
//   SX_CHECK_VAR (mixFactor >= 0. && mixFactor <= 1., mixFactor);
//
//   for (int iSpin=0; iSpin < control->nSpin; iSpin++)  {
//      rhoR(iSpin) = rhoR(iSpin)  *       mixFactor
//                  + rhoIn(iSpin) * (1. - mixFactor);
//   }
//
//   return *this;
//}



void SxRho::readRho (const SxBinIO &io)
{
   SX_CHECK (rBasisPtr);
   const SxRBasis &R  = *rBasisPtr;

   int meshSize = R.getMeshSize ();
   const SxVector3<Int> mesh = R.getMesh ();

   int iSpin, nSpin = getNSpin ();

   try  {
      SxMatrix3<Double> cell;
      SxVector3<Int>    dim;
      rhoR = io.readMesh (&cell, &dim);
      if (nSpin <= 0) nSpin = int(rhoR.getSize ());
      for (iSpin=0; iSpin < nSpin; iSpin++)  rhoR(iSpin).setBasis (&R);


      if (rhoR.getSize() != nSpin)  {
         cout << SX_SEPARATOR;
         sxprintf ("| Charge density file does not contain the correct number "
                 "of spin channels!\n");
         sxprintf ("| The file contains %d spin components whereas %d are "
                 "expected.\n", int(rhoR.getSize()), nSpin);
         cout << SX_SEPARATOR;
         SX_EXIT;
      }

      if (mesh(0) != dim(0) || mesh(1) != dim(1) || mesh(2) != dim(2))
      {
         cout << SX_SEPARATOR;
         sxprintf ("| The mesh size %dx%dx%d of the input density is wrong.\n",
                  dim(0), dim(1), dim(2));
         sxprintf ("| (should be %dx%dx%d)\n", mesh(0), mesh(1), mesh(2));
         cout << SX_SEPARATOR;
         SX_EXIT;
      }

      if (rhoR(0).getSize() != meshSize)  {
         cout << SX_SEPARATOR;
         sxprintf ("| The mesh size of the input density file is wrong!\n");
         sxprintf ("| The input density has %d elements (%d expected).\n", 
                 (int)rhoR(0).getSize(), meshSize);
         cout << SX_SEPARATOR;
         SX_EXIT;
      }

      if ( (cell - R.cell).absSqr().sum() > 1e-5 )  {
         cout << SX_SEPARATOR;
         sxprintf ("| The geometry of the input density is wrong!\n");
         sxprintf ("| Input density: [[%g,%g,%g],[%g,%g,%g],[%g,%g,%g]]\n",
                 cell(0,0), cell(0,1), cell(0,2),
                 cell(1,0), cell(1,1), cell(1,2),
                 cell(2,0), cell(2,1), cell(2,2));
         sxprintf ("| should be:     [[%g,%g,%g],[%g,%g,%g],[%g,%g,%g]]\n",
                 R.cell(0,0), R.cell(0,1), R.cell(0,2),
                 R.cell(1,0), R.cell(1,1), R.cell(1,2),
                 R.cell(2,0), R.cell(2,1), R.cell(2,2));
         cout << SX_SEPARATOR;
         SX_EXIT;
      }

   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }
}


void SxRho::writeRho (SxBinIO &io) const
{
   try  {
      io.writeMesh (rhoR, rBasisPtr->cell, rBasisPtr->getMesh ());
   } catch (SxException e)  {
      e.print ();
      SX_EXIT;
   }
}

// --- SxDensity interface
void SxRho::operator= (const SxDensity &in)
{
   const SxRho &inRef = in.getRef<SxRho> ();
   rhoR.resize (inRef.rhoR.getSize ());
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)
      rhoR(iSpin) = inRef.rhoR(iSpin);
   if (!rBasisPtr) rBasisPtr = inRef.rBasisPtr;
}

void SxRho::operator+= (const SxDensity &x)
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VARS (rhoR.getSize () == xRef.rhoR.getSize (),
                  rhoR.getSize (), xRef.rhoR.getSize ());
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)
      rhoR(iSpin) += xRef.rhoR(iSpin);
}

void SxRho::operator-= (const SxDensity &x)
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VARS (rhoR.getSize () == xRef.rhoR.getSize (),
                  rhoR.getSize (), xRef.rhoR.getSize ());
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)
      rhoR(iSpin) -= xRef.rhoR(iSpin);
}

void SxRho::plus_assign_ax (double a, const SxDensity &x)
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VARS (rhoR.getSize () == xRef.rhoR.getSize (),
                  rhoR.getSize (), xRef.rhoR.getSize ());
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)
      rhoR(iSpin).plus_assign_ax(a, xRef.rhoR(iSpin));
}

void SxRho::plus_assign_aspin (double a, const SxDensity &x)
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VAR (getNSpin () == 2, getNSpin ());
   SX_CHECK_VAR (xRef.getNSpin () == 1, xRef.getNSpin ());
   rhoR(0).plus_assign_ax( 0.5 * a, xRef.rhoR(0));
   rhoR(1).plus_assign_ax(-0.5 * a, xRef.rhoR(0));
}

/// Scalar product
double SxRho::operator| (const SxDensity &x) const
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VARS (rhoR.getSize () == xRef.rhoR.getSize (),
                  rhoR.getSize (), xRef.rhoR.getSize ());
   double res = 0.;
   //const SxGBasis &G = rBasisPtr->getGBasis ();
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)  {
      res += dot((*this)(iSpin), xRef(iSpin));
      // PsiG xG = G | xRef(iSpin), myG = G | ((*this)(iSpin));
      // res += 10. * dot (myG, (1. - G.g2/(2. + G.g2) ) * xG).re; 
   }
   return res / (double)rhoR(0).getSize ();
}

double SxRho::normSqr () const
{
   double res = 0.;
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)
      res += rhoR(iSpin).normSqr ();
   return res / (double)rhoR(0).getSize ();
}

SxDensity SxRho::operator- (const SxDensity &x) const
{
   const SxRho &xRef = x.getRef<SxRho> ();
   SX_CHECK_VARS (rhoR.getSize () == xRef.rhoR.getSize (),
                  rhoR.getSize (), xRef.rhoR.getSize ());
   SxPtr<SxRho> res = SxPtr<SxRho>::create ();
   if (rBasisPtr)
      res->rBasisPtr = rBasisPtr;
   else
      res->rBasisPtr = xRef.rBasisPtr;
   int nSpin = getNSpin ();
   res->rhoR.resize (nSpin);
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)
      res->rhoR(iSpin) = rhoR(iSpin) - xRef.rhoR(iSpin);
   return SxDensity(res);
}

SxDensity SxRho::getCopy () const
{
   SxPtr<SxRho> res = SxPtr<SxRho>::create ();
   res->rBasisPtr = rBasisPtr;
   res->nElectrons = nElectrons;
   int nSpin = getNSpin ();
   res->rhoR.resize (nSpin);
   for (int iSpin = 0; iSpin < nSpin; ++iSpin)
      res->rhoR(iSpin).copy(rhoR(iSpin));
   return SxDensity(res);
}

SxDensity SxRho::spin () const
{
   SX_CHECK_VAR (rhoR.getSize () == 2, rhoR.getSize ());
   SxPtr<SxRho> res = SxPtr<SxRho>::create ();
   res->rBasisPtr = rBasisPtr;
   res->nElectrons = -1.;
   res->rhoR.resize (1);
   res->rhoR(0) = rhoR(0) - rhoR(1);
   return SxDensity(res);
}

void SxRho::renormalize ()
{
   /*
   int nSpin = rhoR.getSize ();
   double norm = 0.;
   for (int iSpin=0; iSpin < nSpin; iSpin++)
      norm += rhoR(iSpin).sum ();
   norm *= dOmega;
   if (fabs (norm - nElectrons) > 1e-8 * nElectrons)  {
      double adjustRho = (nElectrons - norm) / (dOmega * nSpin);
      for (iSpin=0; iSpin < nSpin; iSpin++)
         rhoR(iSpin) += adjustRho;
   }
   */
   normalizeRho ();
}

void SxRho::syncMPI ()
{
#ifdef USE_LOOPMPI
   for (int iSpin = 0; iSpin < rhoR.getSize (); ++iSpin)  {
      if (rBasisPtr)
         rhoR(iSpin).resize (rBasisPtr->getNElements ());
      SxLoopMPI::bcast (rhoR(iSpin), 0);
      VALIDATE_VECTOR(rhoR(iSpin));
   }
#endif
}

