#include <SxFileIO.h>
#include <SxFSAction.h>

#ifdef WIN32
#  include <windows.h>
#else
#  include <dirent.h>
#endif /* WIN32 */

#include <errno.h>


void SxFileIO::syncDir (const SxString &path_)
{
   SX_TRACE ();
#  ifndef WIN32
      DIR *dir = opendir (path_.ascii ());
      if (!dir)
         SX_THROW (SxFileIO::dirOpenErr (path_));

      // or int fd = open (path_.ascii (), O_RDONLY);
      int fd = dirfd (dir);
      if (fd < 0)
         SX_THROW (SxFileIO::dirFDErr (path_));

      if (fsync (fd) != 0)
         SX_THROW (SxFileIO::dirSyncErr (path_));

      if (closedir (dir) != 0)
         SX_THROW (SxFileIO::dirCloseErr (path_));
#  endif /* WIN32 */
}
// --------------------------------------------------------------------------

SxFileIO::SxFileIO ()
   : fp(NULL),
     fileMode(0),
     fileSize(0),
     fileOffset(0),
     path(""),
     lastBytesRead(0),
     lastBytesWritten(0)
{
   SX_TRACE ();
}

SxFileIO::~SxFileIO ()
{
   SX_TRACE ();
   close ();
}

void SxFileIO::enableMode (unsigned int mode_)
{
   SX_TRACE ();
   fileMode |= mode_;
}

void SxFileIO::disableMode (unsigned int mode_)
{
   SX_TRACE ();
   fileMode &= ~(mode_);
}

bool SxFileIO::hasMode (enum SxFileIO::Mode mode_) const
{
   SX_TRACE ();
   return (fileMode & static_cast<unsigned int>(mode_));
}

void SxFileIO::setModeFlags ()
{
   SX_TRACE ();
   SxString m;
   if (hasMode (SxFileIO::ModeAppend)) {
      m += "a";
      if (hasMode (SxFileIO::ModeRead))
         m += "+";
      if (hasMode (SxFileIO::ModeBinary))
         m += "b";
      modeStr = m;
      return;

   } else if (hasMode (SxFileIO::ModeRead) &&
              !hasMode (SxFileIO::ModeCreate)) {
      m += "r";
      if (hasMode (SxFileIO::ModeWrite))
         m += "+";
      if (hasMode (SxFileIO::ModeBinary))
         m += "b";
      modeStr = m;
      return;

   } else if (hasMode (SxFileIO::ModeWrite)) {
      m += "w";
      if (hasMode (SxFileIO::ModeRead))
         m += "+";
      if (hasMode (SxFileIO::ModeBinary))
         m += "b";
      modeStr = m;
      return;
   }
}

void SxFileIO::setSize () 
{
   SX_TRACE ();
   int64_t lastOffset = fileOffset;
   fSeek (0, SEEK_END);
   fileSize = fTell ();
   fSeek (lastOffset, SEEK_SET);
   fileOffset = lastOffset;
}

void SxFileIO::open (const SxString &path_,
                     const SxString &mode_,
                     int            permission_)
{
   SX_TRACE ();
   SX_DBG_MSG ("path='" << path_
               << "', mode='" << mode_
               << "', permission=" << SxString::sprintf("%04o", permission_));

   close ();
   path = path_;

   // --- get the filemode
   fileMode = 0;
   uint64_t i = 0;
   uint64_t n = mode_.getSize ();
   while (i < n)  {
      char c = mode_(i);
      if (c == 'r') {
         if (i + 1 < n && mode_(i + 1) == '+') {
            enableMode ((SxFileIO::ModeRead | SxFileIO::ModeWrite));
            i++;
         } else {
            enableMode (SxFileIO::ModeRead);
         }
      } else if (c == 'w')  {
         if (i + 1 < n && mode_(i + 1) == '+') {
            enableMode ((SxFileIO::ModeRead  |
                         SxFileIO::ModeWrite |
                         SxFileIO::ModeCreate));
            i++;
         } else {
            enableMode (SxFileIO::ModeWrite |
                        SxFileIO::ModeCreate);
         }
      } else if (c == 'a') {
         if (i + 1 < n && mode_(i + 1) == '+') {
            enableMode (SxFileIO::ModeRead   |
                        SxFileIO::ModeWrite  |
                        SxFileIO::ModeAppend |
                        SxFileIO::ModeCreate);
            i++;
         } else {
            enableMode (SxFileIO::ModeWrite  |
                        SxFileIO::ModeAppend |
                        SxFileIO::ModeCreate);
         }
      } else if (c == 'b') {
         enableMode (SxFileIO::ModeBinary);
      } else {
         SX_EXIT; // invalid mode
      }
      i++;
   }
   setModeFlags ();
   // --- create file if it does not exists
   SxFileInfo f(path);
   path = f.getAbsPath ();
   if (!f.exists ()) {
      if (!hasMode (SxFileIO::ModeCreate))
         SX_THROW (SxFileIO::noSuchFileErr (path));

      // --- create new file in corresponding mode   
      fp = fopen (path.ascii (), "w+b");
      if (!fp) {
         SX_THROW (SxFileIO::fileCreateErr (path));
      }
      if (fclose (fp) != 0) {
         fp = NULL;
         SX_THROW (SxFileIO::fileCloseErr (path));
      }
      fp = NULL;
      SxFSAction::chmod (path, permission_);
   }
   
   // --- open file in the correct mode
   fp = fopen (path.ascii (), modeStr.ascii ());
   if (!fp) {
      SX_THROW (SxFileIO::fileOpenErr (path));
   }
   setSize ();
   tell (); // fetch/set current file pointer
   return;
}

SxArray<char> SxFileIO::readLines (int64_t nLines)
{
   SX_TRACE ();
   SX_CHECK (isOpen () && !hasMode (SxFileIO::ModeBinary));

   if (nLines <= 0) {
      nLines = fileSize; // max lines possible
   }

   if (fileSize <= 0) return SxArray<char> ();

   int64_t count = 0;
   int64_t pos = 0;

   SxArray<char> buffer (fileSize);
   int ch;
   while ((ch = fgetc (fp)) != EOF) {
      if ((char)ch == '\0') // ignore null characters
         continue;
      buffer (pos++) = (char)ch;
      if ((char)ch == '\n')  count++;
      if (count == nLines)  break;

   }
   if (ferror (fp)) {
      SX_THROW ("FileReadError", SxFileIO::fileReadErr (path));
   }
   return SxArray<char> (buffer.elements, pos);
}

SxArray<char> SxFileIO::readTail (int64_t nLines)
{
   SX_TRACE ();
   SX_CHECK (isOpen () && !hasMode (SxFileIO::ModeBinary));

   if (nLines <= 0) {
      nLines = fileSize; // max lines possible
   }

   if (fileSize <= 0) return SxArray<char> ();

   int64_t count = 0;
   int64_t pos = 0;
   SxArray<char> buffer (fileSize);

   int ch;
   while ((ch = fgetc (fp)) != EOF) {
      if ((char)ch == '\0') // ignore null characters
         continue;
      if ((char)ch == '\n')  count++;
   }

   if (ferror (fp)) {
      SX_THROW ("FileReadError", SxFileIO::fileReadErr (path));
   }

   seek (0, Seek::BEG);
   int64_t tmpLines = 0;
   SxString res;
   ch = 0;
   while ((ch = fgetc (fp)) != EOF) {
      if ((char)ch == '\0')
         continue;
      if ((char)ch == '\n')  tmpLines++;
      if (tmpLines >= (count - nLines) && tmpLines != count ) {
         buffer (pos++) = (char)ch;
      }
   }
   if (ferror (fp)) {
      SX_THROW ("FileReadError", SxFileIO::fileReadErr (path));
   }
   return SxArray<char> (buffer.elements, pos);
}

uint64_t SxFileIO::read (SxString *str)
{
   SX_TRACE ();
   SX_CHECK (isOpen ());
   SX_CHECK (fileSize > 0 && fileOffset < fileSize);
   SX_DBG_MSG ("size='" << str->getNBytes () << "'");

   uint64_t nBytes = str->getNBytes ();

   uint64_t nBytesLeft = fileSize - fileOffset;
   uint64_t nBytesToRead = nBytesLeft < nBytes ? nBytesLeft : nBytes;

   if (nBytes == 0 || nBytesLeft <= 0) return 0;

   SxArray<char> buf(nBytesToRead);

   if (hasMode (SxFileIO::ModeRead)) {

      int64_t nBytesRead = fread (buf.elements, 1, nBytesToRead, fp);
      lastBytesRead = (nBytesRead > 0)? nBytesRead : 0;
      if (nBytesRead != (int64_t)nBytesToRead)
         SX_THROW (SxFileIO::fileReadErr (path));
      (*str) = SxString(buf.elements, nBytesRead);
      fileOffset += lastBytesRead;
      return (uint64_t)nBytesRead;

   } else { SX_EXIT; } // invalid mode
}

uint64_t SxFileIO::read (void *outPtr, uint64_t nBytes)
{
   SX_TRACE ();
   SX_CHECK (isOpen ());
   SX_CHECK (fileSize > 0 && fileOffset < fileSize);
   SX_DBG_MSG ("size='" << nBytes << "'");

   uint64_t nBytesLeft = fileSize - fileOffset;
   uint64_t nBytesToRead = nBytesLeft < nBytes ? nBytesLeft : nBytes;

   if (nBytes == 0 || nBytesLeft <= 0) return 0;

   if (hasMode (SxFileIO::ModeRead)) {

      int64_t nBytesRead = fread (outPtr, 1, nBytesToRead, fp);
      lastBytesRead = (nBytesRead > 0)? nBytesRead : 0;
      if (nBytesRead != (int64_t)nBytesToRead)
         SX_THROW (SxFileIO::fileReadErr (path));
      fileOffset += lastBytesRead;
      return (uint64_t)nBytesRead;

   } else { SX_EXIT; } // invalid mode
}

uint64_t SxFileIO::write (const SxString &str)
{
   SX_TRACE ();
   SX_DBG_MSG ("size='" << str.getSize () << "'");
   SX_CHECK (isOpen ());

   uint64_t nBytesToWrite = str.getNBytes ();

   if (hasMode(SxFileIO::ModeWrite) || hasMode(SxFileIO::ModeAppend)) {

      int64_t nBytesWritten = fwrite (str.getElems (), 1, nBytesToWrite, fp);
      lastBytesWritten = (nBytesWritten > 0)? nBytesWritten : 0;
      if (nBytesWritten != (int64_t)nBytesToWrite)
         SX_THROW ("FileWriteError", SxFileIO::fileWriteErr (path));
      fileSize += lastBytesWritten;
      return (uint64_t)nBytesWritten;

   } else { SX_EXIT; } // invalid mode
}

uint64_t SxFileIO::write (const void *inPtr, uint64_t nBytes)
{
   SX_TRACE ();
   SX_DBG_MSG ("size='" << nBytes << "'");
   SX_CHECK (isOpen ());

   if (hasMode(SxFileIO::ModeWrite) || hasMode(SxFileIO::ModeAppend)) {

      int64_t nBytesWritten = fwrite (inPtr, 1, nBytes, fp);
      lastBytesWritten = (nBytesWritten > 0)? nBytesWritten : 0;
      if (nBytesWritten != (int64_t)nBytes)
         SX_THROW ("FileWriteError", SxFileIO::fileWriteErr (path));
      fileSize += lastBytesWritten;
      return (uint64_t)nBytesWritten;

   } else { SX_EXIT; } // invalid mode
}

void SxFileIO::cleanup ()
{
   SX_TRACE ();
   SX_CHECK (fp != NULL);
   fp = NULL;
   path = "";
   modeStr = "";
   fileMode = 0;
   fileSize = 0;
   fileOffset = 0;
   lastBytesRead = 0;
   lastBytesWritten = 0;
}

void SxFileIO::close ()
{
   SX_TRACE ();
   if (!isOpen()) return;
   SX_DBG_MSG ("path='" << path << "'");
   if (fclose (fp) != 0) {
      cleanup ();
      SX_THROW (SxFileIO::fileCloseErr (path));
   }
   cleanup ();
   //path = ""; keep for error messages
}

bool SxFileIO::isOpen () const
{
   SX_TRACE ();
   return (bool)(fp);
}

bool SxFileIO::isOpenForWriting () const
{
   SX_TRACE ();
   return (isOpen () && hasMode (SxFileIO::ModeWrite));
}

uint64_t SxFileIO::getSize () const
{
   SX_TRACE ();
   return fileSize;
}

uint64_t SxFileIO::getOffset () const
{
   SX_TRACE ();
   return fileOffset;
}

uint64_t SxFileIO::getLastRead () const
{
   SX_TRACE ();
   return lastBytesRead;
}

SxString SxFileIO::getFilePath () const
{
   SX_TRACE ();
   return path;
}

uint64_t SxFileIO::getLastWritten () const
{
   SX_TRACE ();
   return lastBytesWritten;
}

void SxFileIO::fFlush ()
{
   SX_TRACE ();
   if (fp) {
      if (fflush (fp) != 0)
         SX_THROW (SxFileIO::fileFlushErr (path));
   }
}

void SxFileIO::flush ()
{
   SX_TRACE ();
   SX_CHECK (isOpen ());

   fFlush ();
}

uint64_t SxFileIO::fTell ()
{
   SX_TRACE ();
   int64_t offset = 0;
#  ifdef WIN32
      if ((offset = _ftelli64 (fp)) < 0) {
         SX_THROW (SxFileIO::fileTellErr (path));
      }
#  else
      if ((offset = ftello64 (fp)) < 0) {
         SX_THROW (SxFileIO::fileTellErr (path));
      }
#endif
   return (uint64_t)offset;
}

uint64_t SxFileIO::tell ()
{
   SX_TRACE ();
   SX_CHECK (isOpen ());
   fileOffset = fTell ();
   return fileOffset;
}

void SxFileIO::fSeek (int64_t offset, int origin)
{
   SX_TRACE ();
#if defined(WIN32)
   if (_fseeki64 (fp, offset, origin) != 0)
      SX_THROW (SxFileIO::fileSeekErr (path));
#elif defined(LINUX)
   if (fseeko64 (fp, offset, origin) != 0)
      SX_THROW (SxFileIO::fileSeekErr (path));
#elif defined(MACOSX)
   if ((fileOffset = fseeko (fp, offset, origin)) != 0)
      SX_THROW (SxFileIO::fileSeekErr (path));
#else
#   error "fseek for LFS not yet ported."
#endif
}

void SxFileIO::seek (int64_t offset, SxFileIO::Seek from)
{
   SX_TRACE ();
   SX_CHECK (isOpen ());
   if (offset > (int64_t)fileSize)
      SX_THROW ("Can't set file '"+path+"' position to given offset ");

   fSeek (offset, (int)from);
   fileOffset = fTell ();
}

// --------------------------------------------------------------------------
SxString SxFileIO::noSuchFileErr (const SxString &path)
{
   return "Can't open file '"+path+"': No such file";
}

SxString SxFileIO::fileModeErr (const SxString &path,
                                const SxString &modes)
{
   return "Can't open file '"+path+"': File modes '" + modes
          + "' are mutually exclusive.";
}

SxString SxFileIO::fileCreateErr (const SxString &path)
{
   return "Can't create file '"+path+"': " + sxstrerror ();
}

SxString SxFileIO::fileOpenErr (const SxString &path)
{
   return "Can't open file '"+path+"': " + sxstrerror ();
}

SxString SxFileIO::fileCloseErr (const SxString &path)
{
   return "Can't close file '"+path+"': " + sxstrerror ();
}

SxString SxFileIO::fileReadErr (const SxString &path)
{
   return "Read error in file '"+path+"': " + sxstrerror ();
}

SxString SxFileIO::fileReadBufferErr (const SxString &path)
{
   return "Read error in file '"+path+
          "': Buffer size is smaller than requested nElems to read";
}

SxString SxFileIO::fileWriteErr (const SxString &path)
{
   return "Write error in file '"+path+"': " + sxstrerror ();
}

SxString SxFileIO::fileWriteBufferErr (const SxString &path)
{
   return "Write error in file '"+path+
          "': Buffer size is smaller than requested nElems to write";
}

SxString SxFileIO::fileSeekErr (const SxString &path)
{
   return "Can't set file '"+path+"' position to given offset: " + sxstrerror ();
}

SxString SxFileIO::fileTellErr (const SxString &path)
{
   return "Can't get file '"+path+"' position: " + sxstrerror ();
}

SxString SxFileIO::fileFlushErr (const SxString &path)
{
   return "Can't flush file '" + path + "': " + sxstrerror ();
}

SxString SxFileIO::fileSyncErr (const SxString &path)
{
   return "Can't sync file '" + path + "': " + sxstrerror ();
}

SxString SxFileIO::dirOpenErr (const SxString &path)
{
   return "Can't open directory '" + path + "': " + sxstrerror ();
}

SxString SxFileIO::dirFDErr (const SxString &path)
{
   return "Can't get descriptor from directory '" + path + "': " + sxstrerror ();
}

SxString SxFileIO::dirSyncErr (const SxString &path)
{
   return "Can't sync directory '" + path + "': " + sxstrerror ();
}

SxString SxFileIO::dirCloseErr (const SxString &path)
{
   return "Can't close directory '" + path + "': " + sxstrerror ();
}

// --- test functions

void SxFileIO::printModeStr () const
{
   sxprintf ("Mode string: %s\n", modeStr.ascii ());
}

// --- Streambuf functions

SxFileIO::int_type SxFileIO::overflow (int c)
{
   return sputc (static_cast<char>(c));
}

SxFileIO::int_type SxFileIO::sputc (char c)
{
   write ((void *)(&c), 1);
   return c;
}

SxFileIO::int_type SxFileIO::uflow ()
{
   return sgetc ();
}

SxFileIO::int_type SxFileIO::underflow ()
{
   int_type c = sgetc ();
   seek ((--fileOffset), Seek::BEG);
   return c;
}

SxFileIO::int_type SxFileIO::sgetc ()
{
   if (fileOffset >= fileSize)
      return EOF;
   char c;
   read (c);
   return c;
}
