#ifndef _SX_FILE_IO_H_
#define _SX_FILE_IO_H_

#include <SxFS.h>
#include <SxString.h>
#include <SxError.h>
#include <streambuf>

#ifdef WIN32
#   include <windows.h>
#else
#   define _FILE_OFFSET_BITS 64
#   include <unistd.h>
#endif


/** \brief File I/O

   implemented: - mode 'f'
                - sync // only flushes for fileStream mode
                - syncDir // for linux


   \par Overview
   This class hides file implementation differences on Linux/Mac/Windows.
   Synchronous write on Windows requires different file handle and functions
   than on POSIX systems. 

   \par Directory synchronization (syncDir)
   Synchronize directory to hard drive after creating or removing a file or
   a subdirectory.

   \par OS
   Different implementation for POSIX and Windows. Small details for MacOS.

   \par Open File Modes
\verbatim
       Open for             !Existing file   Stream position
       -------------------  ---------------  ---------------
   r   reading,                              beginning
   w   writing,             create/truncate, beginning
   a   writing,             created        , end
   r+  reading and writing,                , beginning
   w+  reading and writing, create/truncate, beginning
   a+  reading and writing, created        , end

   b   binary
\endverbatim
 */
class SX_EXPORT_FS SxFileIO : public std::streambuf
{
   public:
      enum Mode {
         ModeUndefined  = 0,
         ModeRead       = (1 << 0), // r, r+, w+, a+
         ModeWrite      = (1 << 1), // r+, w, w+, a, a+
         ModeAppend     = (1 << 2), // a, a+
         ModeCreate     = (1 << 3), // w, w+, a, a+
         ModeBinary     = (1 << 4), // b
      };

      enum class Seek
      {
         BEG = SEEK_SET,
         CUR = SEEK_CUR,
         END = SEEK_END
      };

      SxFileIO ();
     ~SxFileIO ();

      void open (const SxString &path,
                 const SxString &mode="",
                 int            permission=0600);

      void close ();

      bool isOpen () const;

      bool isOpenForWriting () const;

      bool hasMode (enum Mode) const;

      uint64_t getSize () const;
      uint64_t getOffset () const;

      uint64_t getLastRead () const;
      uint64_t getLastWritten () const;

      SxString getFilePath () const;

      template<typename... Args>
      int printf (const char *format, Args... args);
      template<typename... Args>
      int scanf (const char *format, Args... args);

      template<typename T>
      uint64_t read (SxArray<T> *outBufPtr, int64_t nElems = -1);
      uint64_t read (SxString *str);
      uint64_t read (void *outPtr, uint64_t nBytes);
      template<typename T>
      uint64_t read (T &obj);

      SxArray<char> readLines (int64_t nLines = 0);
      SxArray<char> readTail (int64_t nLines = 0);

      template<typename T>
      uint64_t write (const SxArray<T> &buffer, int64_t nElems = -1);
      uint64_t write (const SxString &str);
      uint64_t write (const void *inPtr, uint64_t nBytes);
      template<typename T>
      uint64_t write (const T &obj);


      uint64_t tell ();
      void seek (int64_t offset, SxFileIO::Seek from = Seek::CUR);

      // --- flushes the application buffers
      void flush ();

      // syncDir in order to update directory after file creation
      static void syncDir (const SxString &path);

      // --- test functions
      void printModeStr () const;

      // static read/write functions
      static inline SxArray<char> readLines (const SxString &filename,
                                             int64_t nLines=0);
      static inline SxArray<char> readBinary (const SxString &filename,
                                              int64_t nBytes);
      static inline SxArray<char> readTail (const SxString &filename,
                                            int64_t nLines = 0);

      static inline void write (const SxString &str, const SxString &filename,
                                int mode);
      static inline void appendToFile (const SxString &str,
                                       const SxString &filename);


      operator istream& () {
         if (istreamPtr.getPtr () == NULL)
            istreamPtr = SxPtr<istream>::create (static_cast<streambuf*>(this));
         return *istreamPtr;
      }

      operator ostream& () {
         if (ostreamPtr.getPtr () == NULL)
            ostreamPtr = SxPtr<ostream>::create (static_cast<streambuf*>(this));
         return *ostreamPtr;
      }

      template<typename T>
      friend SxFileIO &operator<< (SxFileIO &f, const SxArray<T> &buffer) {
         try {
            f.write (buffer, buffer.getSize());
         } catch (SxException ex) {
            SX_THROW (ex, "FileWriteError",
                      SxFileIO::fileReadErr (f.getFilePath ()));
         }
         return f;
      }

      template<typename T>
      friend SxFileIO &operator>> (SxFileIO &f, SxArray<T> &buffer) {
         try {
            f.read (&buffer, buffer.getSize());
         } catch (SxException ex) {
            SX_THROW (ex, "FileReadError",
                      SxFileIO::fileWriteErr (f.getFilePath ()));
         }
         return f;
      }

   private:

      SxPtr<ostream> ostreamPtr;
      SxPtr<istream> istreamPtr;

   protected:


      // --- stream mode
      FILE *fp;

      void enableMode (unsigned int);
      void disableMode (unsigned int);
      void setModeFlags ();
      unsigned fileMode;
      SxString modeStr;    // cstdlib fopen ()

      void setSize ();
      uint64_t fileSize;
      uint64_t fileOffset;

      SxString path;

      uint64_t lastBytesRead;
      uint64_t lastBytesWritten;


      // --- file stream related
      uint64_t fTell ();
      void fSeek (int64_t offset, int origin);
      void fFlush ();

      void cleanup ();

      static SxString noSuchFileErr (const SxString &path);
      static SxString fileModeErr (const SxString &path,
                                   const SxString &modes);
      static SxString fileCreateErr (const SxString &path);
      static SxString fileOpenErr (const SxString &path);
      static SxString fileCloseErr (const SxString &path);
      static SxString fileReadErr (const SxString &path);
      static SxString fileReadBufferErr (const SxString &path);
      static SxString fileWriteErr (const SxString &path);
      static SxString fileWriteBufferErr (const SxString &path);
      static SxString fileSeekErr (const SxString &path);
      static SxString fileTellErr (const SxString &path);

      static SxString fileFlushErr (const SxString &path);
      static SxString fileSyncErr (const SxString &path);
      static SxString dirOpenErr (const SxString &path);
      static SxString dirFDErr (const SxString &path);
      static SxString dirSyncErr (const SxString &path);
      static SxString dirCloseErr (const SxString &path);


      // --- streambuf functions
      SxFileIO::int_type overflow (int_type c = EOF);
      SxFileIO::int_type sputc (char c);
      SxFileIO::int_type uflow ();
      SxFileIO::int_type underflow ();
      SxFileIO::int_type sgetc ();

};

SxArray<char> SxFileIO::readLines (const SxString &filename, int64_t nLines)
{
   SX_TRACE ();

   SxFileIO f;
   f.open (filename.ascii (), "r");
   return f.readLines (nLines); 
}

SxArray<char> SxFileIO::readBinary (const SxString &filename,
                                    int64_t nBytesToRead)
{
   SX_TRACE ();

   SxFileIO f;
   f.open (filename.ascii (), "rb");
   uint64_t fSize = f.getSize ();
   if (nBytesToRead < 0) nBytesToRead = fSize;
   if ((uint64_t)nBytesToRead > fSize) {
      f.close ();
      SX_THROW ("Can't read " + SxString(nBytesToRead) + " bytes from file '"
                 + filename+"': File size is only "+SxString(fSize)+" bytes.");
   }
   SxArray<char> buffer (nBytesToRead);

   if (nBytesToRead == 0)  return SxArray<char>();

   uint64_t nBytesRead = f.read (&buffer, nBytesToRead);
   f.close ();
   return SxArray<char> (buffer.elements, nBytesRead);
}

SxArray<char> SxFileIO::readTail (const SxString &filename, int64_t nLines)
{
   SX_TRACE ();

   SxFileIO f;
   f.open (filename.ascii (), "r");
   return f.readTail (nLines);
}

void SxFileIO::write (const SxString &str, const SxString &filename,
                      int permissions)
{
   SX_TRACE ();

   const int minPerm = 0;
   const int maxPerm = 7777;

   if (permissions < minPerm || permissions > maxPerm)  {
      SX_THROW ("Can't open file '" + filename + "' for writing "
                "with permission mode 0" + SxString::sprintf("%04o", permissions)
                );
   }

   SxFileIO f;
   f.open (filename.ascii (), "wb", permissions);

   uint64_t nBytesToWrite = str.getNBytes ();

   SxArray<char> buffer (str.ascii (), nBytesToWrite);

   uint64_t nBytesWritten = f.write (buffer, nBytesToWrite);
   if (nBytesWritten < nBytesToWrite) {
      f.close ();
      SX_THROW ("FileWriteError", "Can't write" + SxString(nBytesToWrite) +
                "to file: "+filename);
   }
   f.close ();
}

void SxFileIO::appendToFile (const SxString &str, const SxString &filename)
{
   SxFileIO f;
   f.open (filename.ascii (), "a");

   uint64_t nBytesToWrite = str.getNBytes ();

   SxArray<char> buffer (str.ascii (), nBytesToWrite);

   uint64_t nBytesWritten = f.write (buffer, nBytesToWrite);
   if (nBytesWritten < nBytesToWrite) {
      f.close ();
      SX_THROW ("FileWriteError", "Can't append" + SxString(nBytesToWrite) +
                "to file: "+filename);
   }
   f.close ();
}

template<typename... Args>
int SxFileIO::printf (const char *format, Args... args)
{
   SX_TRACE ();
   SX_CHECK (isOpen () && !hasMode (SxFileIO::ModeBinary));

   if (hasMode (SxFileIO::ModeWrite) || hasMode(SxFileIO::ModeAppend)) {
      int res = ::fprintf (fp, format, args...);
      if (ferror (fp)) {
         SX_THROW ("FileWriteError", SxFileIO::fileWriteErr (path));
      }
      setSize ();
      return res;
   } else { SX_EXIT; } // invalid mode
   return 0;
}

template<typename... Args>
int SxFileIO::scanf (const char *format, Args... args)
{
   SX_TRACE ();
   SX_CHECK (isOpen () && !hasMode (SxFileIO::ModeBinary));

   if (hasMode (SxFileIO::ModeRead)) {
      int res = ::fscanf (fp, format, args...);
      if (ferror (fp)) {
         SX_THROW ("FileReadError", SxFileIO::fileReadErr (path));
      }
      tell ();
      return res;
   } else { SX_EXIT; } // invalid mode
}

template<typename T>
uint64_t SxFileIO::read (SxArray<T> *outBufPtr, int64_t nElems)
{
   SX_TRACE ();
   SX_CHECK (isOpen ());
   SX_CHECK (fileSize > 0 && fileOffset < fileSize);
   SX_DBG_MSG ("size='" << nElems * sizeof (T) << "'");

   if (nElems == -1)  nElems = outBufPtr->getSize ();

   constexpr uint64_t typeSize = sizeof (T);
   uint64_t nItemsLeft = (fileSize - fileOffset) / typeSize;
   uint64_t nItemsToRead = nItemsLeft < (uint64_t)nElems ? nItemsLeft : nElems;

   if (nElems == 0 || nItemsLeft <= 0) return 0;

   if (outBufPtr->getSize () < (int64_t)nItemsToRead)
      SX_THROW ("ReadBufferError",
                SxFileIO::fileReadBufferErr (outBufPtr->getSize ()));

   if (hasMode (SxFileIO::ModeRead)) {

      int64_t nItemsRead = fread (outBufPtr->elements, typeSize, nItemsToRead, fp);
      lastBytesRead = (nItemsRead > 0)? nItemsRead * typeSize : 0;
      if (nItemsRead != (int64_t)nItemsToRead)
         SX_THROW (SxFileIO::fileReadErr (path));
      fileOffset += lastBytesRead;
      return (uint64_t)nItemsRead;

   } else { SX_EXIT; } // invalid mode
}

template<typename T>
uint64_t SxFileIO::read (T &obj)
{
   return read (&obj, sizeof(T));
}

template<typename T>
uint64_t SxFileIO::write (const SxArray<T> &buffer, int64_t nElems)
{
   SX_TRACE ();
   SX_DBG_MSG ("size='" << nElems * sizeof (T) << "'");
   SX_CHECK (isOpen ());

   if (nElems == -1)  nElems = buffer.getSize ();

   constexpr uint64_t typeSize = sizeof(T);
   if (buffer.getSize () < nElems)
      SX_THROW ("WriteBufferError",
                SxFileIO::fileWriteBufferErr (buffer.getSize ()));

   if (hasMode(SxFileIO::ModeWrite) || hasMode(SxFileIO::ModeAppend)) {

      int64_t nItemsWritten = fwrite (buffer.elements, typeSize, nElems, fp);
      lastBytesWritten = (nItemsWritten > 0)? nItemsWritten * typeSize : 0;
      if (nItemsWritten != nElems)
         SX_THROW ("FileWriteError", SxFileIO::fileWriteErr (path));
      fileSize += lastBytesWritten;
      return (uint64_t)nItemsWritten;

   } else { SX_EXIT; } // invalid mode
}

template<typename T>
uint64_t SxFileIO::write (const T &obj)
{
   return write (&obj, sizeof(T));
}

#endif /* _SX_FILE_IO_H_ */
