// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxMPIMgr.h>
#include <SxMPI.h>
#include <SxThread.h>
#include <SxTopology.h>
#include <SxString.h>
#include <SxCLI.h>

static void run_MPI_Finalize ()
{
#ifdef USE_MPI
   (void)MPI_Finalize ();
#endif
}

SxMPIHandle::SxMPIHandle (int argc, char **argv)
   : rank(-1), nProcs(-1)
{
#  ifdef USE_MPI
      MPI_Init (&argc, &argv);
      MPI_Comm_size (MPI_COMM_WORLD, &nProcs);
      MPI_Comm_rank (MPI_COMM_WORLD, &rank);
      if (rank != 0) SxCLI::Log::disable ("not MPI master");
      SxCLI::atExit.append (&run_MPI_Finalize); // to be run if SxCLI exits
#  else
      SX_UNUSED (argc);
      SX_UNUSED (argv);
#  endif
}

SxMPIHandle::~SxMPIHandle () 
{ 
#  ifdef USE_MPI
      MPI_Finalize ();
#  endif
}

// --------------------------------------------------------------------------


// construct top level MPI group
SxMPIMgr::SxMPIMgr ()
   : SxProcUnitGroup (),
     hasComm (false)
{
#  ifdef USE_MPI
      MPI_Comm_dup (MPI_COMM_WORLD, &mpiComm);
      MPI_Comm_create_errhandler (SxMPIMgr::errorHandlerCB, &errHandler);
      MPI_Comm_set_errhandler (MPI_COMM_WORLD, errHandler);
#  endif
   init ();
}

// construct any mpi child group
SxMPIMgr::SxMPIMgr (const SxPtr<SxMPIMgr> &parent_) 
   : SxProcUnitGroup (parent_),
     nProcs(-1),
     hasComm (false)
{
#  ifdef USE_MPI
      MPI_Comm_dup (getGlobalPtr()->mpiComm, &mpiComm);
      init ();
      mpiHandle = getGlobalPtr()->mpiHandle;
#  endif
}

SxMPIMgr::~SxMPIMgr ()
{
// MPI_Comm_free (&mpiComm);
}


void SxMPIMgr::init ()
{
#   ifdef USE_MPI
      MPI_Comm_rank (mpiComm, &rank);
      MPI_Comm_size (mpiComm, &nProcs);
#   endif /* USE_MPI */
}


SxPtr<SxMPIMgr> &SxMPIMgr::getGlobalPtr (int argc, char **argv)
{
   static SxPtr<SxMPIMgr> mgr;
   if (argc < 0)  {
      // magic MPI shutdown
      mgr = SxPtr<SxMPIMgr> ();
   }
   if (!mgr.getPtr() )  {
#     ifdef USE_MPI
         // initialize MPI
         SxPtr<SxMPIHandle> mpiHandle = SxPtr<SxMPIHandle>::create (argc, argv);
         int rank   = mpiHandle->rank;
         int nProcs = mpiHandle->nProcs;

         //  sepcify topology
         SxPtr<SxTopology> topo;
         topo = SxPtr<SxTopology>::create (rank, nProcs, SxTopology::FarmerWorker);

         // setup MPI group
         mgr = SxPtr<SxMPIMgr>::create ();
         mgr->mpiHandle = mpiHandle;
         mgr->topology  = topo;

         // init processes
         SxPtr<SxMPI> p;
         for (int i=0; i < nProcs; ++i)  {
            p = SxPtr<SxMPI>::create (mgr->SxProcUnitGroup::getThis ());
            if (i == rank)  
               mgr->curProc = p;
         }
#     else
         SX_UNUSED (argc);
         SX_UNUSED (argv);
#     endif
//    mgr->reparent (mgr->SxProcUnitGroup::getThis());
//    mgr->start ();
   }
   return mgr;
}


SxPtr<SxMPI> SxMPIMgr::getCurProc ()
{
   return SxMPIMgr::getGlobalPtr()->curProc;
}



// --------------------------------------------------------------------


SxArray<SxPtr<SxMPI> > SxMPIMgr::getProcs (bool recursive)
{
   SxList<SxPtr<SxProcUnit> > units = getProcUnits (recursive);
   SxArray<SxPtr<SxMPI> > res (units.getSize());

   SxList<SxPtr<SxProcUnit> >::Iterator unitIt;
   int i=0;
   for (unitIt = units.begin(); unitIt != units.end(); ++unitIt, ++i) {
      res(i) = SxPtr<SxMPI>(*unitIt);
   }
   return res;
}


void SxMPIMgr::start ()
{
   SxList<SxPtr<SxProcUnitGroup> >::Iterator grIt;
   for (grIt = groups.begin(); grIt != groups.end(); ++grIt)  {
      (*grIt)->start ();
   }
   SxArray<SxPtr<SxMPI> > procs = getProcs ();
   for (int i=0; i < procs.getSize(); ++i)  {
      if (topology->isLocalTask (i))  procs(i)->start ();
   }
}

void SxMPIMgr::barrier ()
{
   SxList<SxPtr<SxProcUnitGroup> >::Iterator grIt;
   for (grIt = groups.begin(); grIt != groups.end(); ++grIt)  {
      (*grIt)->barrier ();
   }
   SxArray<SxPtr<SxMPI> > procs = getProcs ();
   for (int i=0; i < procs.getSize(); ++i)  {
      if (topology->isLocalTask (i))  procs(i)->barrier ();
   }
}


SxPtr<SxProcUnitGroup> SxMPIMgr::addGroup (int maxItems)
{
   SX_CHECK (maxItems == 0);  // != 0 not yet implemented
   SxPtr<SxMPIMgr> grp;
   grp = SxPtr<SxMPIMgr>::create (SxProcUnitGroup::getThis());
   grp->topology = topology->clone ();
   groups << grp;
#  ifdef USE_MPI
      int ranks[1];
      ranks[0] = 0;
#  endif /* USE_MPI */
   return grp;
}


SxPtr<SxProcUnitGroup> SxMPIMgr::getTopLevelGroup ()
{
   return SxMPIMgr::getGlobalPtr ();
}

int SxMPIMgr::getRank ()
{
   int r = 0;
#  ifdef USE_MPI
      MPI_Comm_rank (MPI_COMM_WORLD, &r);
#  endif
   return r;
}


int SxMPIMgr::getNProcs ()
{
   int n = 0;
#  ifdef USE_MPI
      MPI_Comm_size (MPI_COMM_WORLD, &n);
#  endif
   return n;
}


int SxMPIMgr::updateProcs ()
{
#  ifdef USE_MPI
      // --- build up new group
      if (procUnits.getSize() > nProcs)  {
         sxprintf ("ERROR: Not enough free processes available.\n");
         sxprintf ("       nProcs = %d, maxProcs = %d.\n",
                   (int)procUnits.getSize(), nProcs);
         SX_CHECK_VARS (procUnits.getSize() <= nProcs,
                        procUnits.getSize(), nProcs);
         SX_QUIT;
      }

      if (procUnits.getSize() == 0)  return rank;

      topology->update (procUnits.getSize());

      // --- specify ranks belonging to new group
      SxArray<int> ranks (procUnits.getSize());
      for (int i=0; i < ranks.getSize(); ++i)  ranks(i) = i;
      
      /*
      MPI_Group worldGroup, mpiGroup;
      MPI_Comm  worldComm = MPI_COMM_WORLD;

      // --- setup group
      // TODO: execute only on all parents of this group
      //SX_EXIT;
      MPI_Comm_group (worldComm, &worldGroup);
      MPI_Group_incl (worldGroup, ranks.getSize(), ranks.elements, &mpiGroup);

      // --- setup communicator
      if (hasComm)  MPI_Comm_free (&mpiComm);

      hasComm = (rank < procUnits.getSize());

      // TODO: to be executed by all members of parent comm
      // MPI_Comm_create (worldComm, mpiGroup, &tempComm);
      // if (iAmInGroup)  mpiComm = tempComm
      //SX_EXIT;
      MPI_Comm_create (worldComm, mpiGroup, &mpiComm);

      // --- clean up
      MPI_Group_free (&mpiGroup);
      MPI_Group_free (&worldGroup);
      */

      return rank;
#  else
      return 0;
#  endif /* USE_MPI */
}


void SxMPIMgr::deregisterProcUnit (const SxPtr<SxProcUnit> &u)
{
   SxProcUnitGroup::deregisterProcUnit (u);
   updateProcs ();
}




// --------------------------------------------------------------------------

#ifdef USE_MPI
void SxMPIMgr::errorHandlerCB (MPI_Comm * /*comm*/, int *err, ...)
{
   int len, rank;
   char buffer[MPI_MAX_ERROR_STRING];

   MPI_Comm_rank (MPI_COMM_WORLD, &rank);
   MPI_Error_string (*err, buffer, &len);
   sxprintf ("ERROR: An MPI error occurred (err=%d)\n", *err);
   sxprintf ("       %s\n", buffer);
   sxprintf ("       Rank = %d\n", rank);
   SX_EXIT;
}

#endif /* USE_MPI */
