// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxSystemThread.h>
#include <SxError.h>
#include <SxTime.h>
#include <unistd.h>
#ifdef WIN32
#  include <process.h>
#endif

SxSystemThread::SxSystemThread ()
  : running (false),
    hasLastException (false),
    hasThreadId (false)
{
   // empty
}



SxSystemThread::~SxSystemThread ()
{
   if (isRunning ())  SX_EXIT;
}

void SxSystemThread::start (Priority priority)
{
   SX_CHECK (!hasThreadId);

   try {
#  ifdef WIN32
      // --- prepare user data
      userData = (void **)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, 
                          sizeof(void *));
      *userData = this;


      // --- create thread
      // Note: Under MingW we have to use _beginthreadex() instead of 
      //       CreateThread() to avoid memleaks!
      threadHandle = (HANDLE)_beginthreadex (
                        NULL, 0, &loopCallback, userData,
                        0, &threadId.id); 
#  else
      int newPriority=0;

      // --- setup thread priority
#     ifndef SX_ANDROID
      if (priority == SameAsParent)  {
         pthread_attr_setinheritsched (&threadAttribs, PTHREAD_INHERIT_SCHED);
      } else  {
         // --- get default schedule policy
         int schedPolicy;
         int error = pthread_attr_getschedpolicy (&threadAttribs, &schedPolicy);
         SX_CHECK (!error);
         schedPolicy = SCHED_RR;
         pthread_attr_setschedpolicy (&threadAttribs, SCHED_RR);
   
         // --- get allowed priority range
         int minPriority = sched_get_priority_min (schedPolicy);
         int maxPriority = sched_get_priority_max (schedPolicy);
         SX_CHECK (minPriority > 0);
         SX_CHECK (maxPriority > 0);

         if      (priority == Idle)     newPriority = minPriority;
         else if (priority == Highest)  newPriority = maxPriority;
         else {
            newPriority = ((maxPriority - minPriority) / maxPriority) * priority
                        +   minPriority;
         
         }

         sched_param schedParam;
         schedParam.sched_priority = newPriority;

         pthread_attr_setinheritsched (&threadAttribs, PTHREAD_EXPLICIT_SCHED);
         pthread_attr_setschedparam (&threadAttribs, &schedParam);
      }
#     endif /* SX_ANDROID */
   

      int err = pthread_create (&threadId.id, NULL,
                                (void *(*)(void *))loopCallback,
                                this);
      if (err)  {
         // TODO: error
         return;
      }
#  endif
   } catch (...) {
      SX_EXIT; // main routine of derived thread class needs try-catch block
   }
   hasThreadId = true;
}

void SxSystemThread::launcher ()
{
   running = true;
   hasLastException = false;
   try {
      main ();
   }  catch (SxException e) {
      hasLastException = true;
      lastException = e;
   }  catch (...) {
      hasLastException = true;
#  ifdef WIN32
      DWORD err = GetLastError ();
      SxString msg = "System thread interrupted unexpectedly: " 
         + sxstrerror (err) 
         + "(" + sxprintf ("0x%x", err) + ")";
#  else
      SxString msg = "System thread interrupted unexpectedly";
#  endif
      lastException = SxException(msg.elements, __FILE__, __LINE__);
   }
   running = false;
}

bool SxSystemThread::hasException () const
{
   return hasLastException;
}

SxException SxSystemThread::getException () const
{
   return lastException;
}

#ifdef WIN32
unsigned __stdcall SxSystemThread::loopCallback (void *userData)
#else
void SxSystemThread::loopCallback (void *userData)
#endif
{
#  ifdef WIN32
      SxSystemThread *obj = *static_cast<SxSystemThread **>(userData);
      SX_CHECK (obj);
      obj->launcher ();
      _endthreadex (0);
      return 0;
#  else
      SxSystemThread *obj = static_cast<SxSystemThread *>(userData);
      SX_CHECK (obj);
      obj->launcher ();
      pthread_exit(0);
#  endif
}


void SxSystemThread::wait ()
{
   if (hasThreadId)  {
#     ifdef WIN32
         if (threadHandle != NULL)  {
            DWORD result = WaitForSingleObject (threadHandle, INFINITE);
            if (result != WAIT_OBJECT_0)   {
               cout << "Wait for single object unexpectedly quit: " << result << endl;
               if (result == WAIT_FAILED)  {
                  DWORD err = GetLastError ();
                  cout << "Wait for tid " + SxString(threadId.id)
                     + " failed: " + sxstrerror (err);
               }
            }
            SX_MUTEX (waitMutex)  {
               if (userData) HeapFree (GetProcessHeap (), 0, userData);
               userData = NULL;
               if (threadHandle) CloseHandle (threadHandle);
               threadHandle = NULL;
            } 
         }
#     else
         pthread_join (threadId.id, NULL);
#     endif
      SX_MUTEX (waitMutex)  {
         hasThreadId = false;
         threadId = SxThreadId ();
      }
   }
}


// void SxSystemThread::terminate ()
// {
//    running = false;
// #  ifdef WIN32
//       _endthreadex (0);
// #  else
//       pthread_exit (0);
// #  endif
// }


bool SxSystemThread::isRunning () const
{
   return running;
}

void SxSystemThread::main ()
{
   SX_EXIT;  // this function is not pure abstract to allow static linkage
}
