// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

// note1:
//    The SxException class is used by almost all other service classes.
//    Hence, here we cannot use those service classes (like SxString, SxArray).
//    That's why here conventional C memory management has to be applied.


#include <SxException.h>
#include <SxString.h>
#include <SxList.h>
#include <SxError.h>

bool SxException::dumpCoreFile = false;


SxException::SxException ()
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (-1),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   message = new char [1];  // see note1
   message[0] = '\0';

   tag = new char [1];  // see note1
   tag[0] = '\0';
}

SxException::SxException (const SxException &in)
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (in.senderLine),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   try {
      set_new_handler(NULL);
      copy (in);
   } catch (std::bad_alloc &excep) {
      SX_UNUSED(excep);
      in.print (true);
      sxOutOfMemoryHandler ();
   }
   set_new_handler(sxOutOfMemoryHandler);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}


SxException::SxException (const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   SX_DBG_MSG ("throwing (" << message_ << ")" << " from " 
                            << senderFile_ << ":" << senderLine_);
   try {
      set_new_handler(NULL);
      init (message_, NULL, senderFile_, senderLine_);
   } catch (std::bad_alloc &excep) {
      SX_UNUSED(excep);
      std::cout << std::endl << message_ << std::endl;
      std::cout << "This Exception was emitted from "
                << senderFile_
                << " (line " << senderLine_ << ")\n";
      sxOutOfMemoryHandler ();
   }
   set_new_handler(sxOutOfMemoryHandler);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}


SxException::SxException (const char *tag_, const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   SX_DBG_MSG ("throwing (" << message_ << ")" << " from " 
                            << senderFile_ << ":" << senderLine_);
   try {
      set_new_handler(NULL);
      init (message_, tag_, senderFile_, senderLine_);
   } catch (std::bad_alloc &excep) {
      SX_UNUSED(excep);
      std::cerr << std::endl << message_ << std::endl;
      std::cerr << "This Exception was emitted from "
                << senderFile_
                << " (line " << senderLine_ << ")\n";
      sxOutOfMemoryHandler ();
   }
   set_new_handler(sxOutOfMemoryHandler);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}



SxException::SxException (const SxException &chain_,
                          const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   SX_DBG_MSG ("Msg: " << message_ << " from " 
                       << senderFile_ << ":" << senderLine_);

   try {
      set_new_handler(NULL);
      next = new SxException();
      next->copy (chain_);
      init (message_, NULL, senderFile_, senderLine_);
   } catch (std::bad_alloc &excep) {
      SX_UNUSED(excep);
      cerr << "Exception Stack: \n";
      const SxException *ptr = chain_.tail;
      int counter = 1;
      int i = 0;
      int strSize = 0;
      while (ptr) {
         cerr << counter++ << ": ";
         if (ptr->message) cerr << ptr->message << " ";
         strSize = static_cast<int>(strlen (ptr->senderFile));
         i = strSize - 1;
         while (i >= 0 && ptr->senderFile[i] != '\\' && 
                ptr->senderFile[i] != '/') i--;
         cerr << "(";
         while ((i+1) < strSize)  { cerr << ptr->senderFile[i+1]; i++; }
         cerr << ":" << ptr->senderLine << ")" << endl;
         ptr = ptr->prev;
      }
      cerr << counter++ << ": ";
      if (message_) cerr << message_ << " ";
      strSize = static_cast<int>(strlen (senderFile_));
      i = strSize - 1;
      while (i >= 0 && senderFile_[i] != '\\' && senderFile_[i] != '/') i--;
      cerr << "(";
      while ((i+1) < strSize)  { cerr << senderFile_[i+1]; i++; }
      cerr << ":" << senderLine_ << ")" << endl;
      sxOutOfMemoryHandler ();
   }
   set_new_handler(sxOutOfMemoryHandler);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}

SxException::SxException (const SxException &chain_,
                          const char *tag_, const char *message_,
                          const char senderFile_[], int senderLine_)
   : message(NULL),
     tag(NULL),
     senderFile (NULL),
     senderLine (senderLine_),
     next(NULL),
     prev(NULL),
     tail(NULL)
{
   SX_DBG_MSG ("Msg: " << message_ << " from " 
                       << senderFile_ << ":" << senderLine_);
   try {
      set_new_handler(NULL);
      next = new SxException();
      next->copy (chain_);
      init (message_, tag_, senderFile_, senderLine_);
   } catch (std::bad_alloc &excep) {
      SX_UNUSED(excep);
      cerr << "Exception Stack: \n";
      const SxException *ptr = chain_.tail;
      int counter = 1;
      int i = 0;
      int strSize = 0;
      while (ptr) {
         cerr << counter++ << ": ";
         if (ptr->tag)  cerr << "[" << ptr->tag << "] ";
         if (ptr->message) cerr << ptr->message << " ";
         strSize = static_cast<int>(strlen (ptr->senderFile));
         i = strSize - 1;
         while (i >= 0 && ptr->senderFile[i] != '\\' && 
                ptr->senderFile[i] != '/') i--;
         cerr << "(";
         while ((i+1) < strSize)  { cerr << ptr->senderFile[i+1]; i++; }
         cerr << ":" << ptr->senderLine << ")" << endl;
         ptr = ptr->prev;
      }
      cerr << counter++ << ": ";
      if (tag_)  cerr << "[" << tag_ << "] ";
      if (message_) cerr << message_ << " ";
      strSize = static_cast<int>(strlen (senderFile_));
      i = strSize - 1;
      while (i >= 0 && senderFile_[i] != '\\' && senderFile_[i] != '/') i--;
      cerr << "(";
      while ((i+1) < strSize)  { cerr << senderFile_[i+1]; i++; }
      cerr << ":" << senderLine_ << ")" << endl;
      sxOutOfMemoryHandler ();
   }
   set_new_handler(sxOutOfMemoryHandler);

   if (dumpCoreFile)  {
      print ();
      fflush (stdout);
      fflush (stderr);
      SX_BREAK;
   }
}


SxException::~SxException ()
{
   destroy ();
}

SxException &SxException::operator= (const SxException &in)
{
   if (&in != this)  {
      copy (in);
   }
   return *this;
} 

void SxException::init (const SxException &in)
{
   init (in.message, in.tag, in.senderFile, in.senderLine);
}

void SxException::init (const char *message_,
                        const char *tag_,
                        const char *senderFile_,
                        int        senderLine_)
{

   if (message_)  {
      if (message != NULL) delete [] message;
      message = new char [strlen (message_)+1];  // see note1
      strcpy (message, message_);
   }
   if (tag_)  {
      if (tag != NULL)  delete [] tag;
      tag = new char [strlen (tag_)+1];  // see note1
      strcpy (tag, tag_);
   }
   if (senderFile_)  {
      if (senderFile != NULL)  delete [] senderFile;
      senderFile = new char [strlen (senderFile_)+1];  // see note1
      strcpy (senderFile, senderFile_);
   }
   senderLine = senderLine_;
}

void SxException::destroy ()
{
   if (message)  {
      delete [] message;
      message = NULL;
   }
   if (tag)  {
      delete [] tag;
      tag = NULL;
   }
   if (senderFile)  {
      delete [] senderFile;
      senderFile = NULL;
   }
   if (next)  {
      delete next;
      next = NULL;
   }

   if (prev) {
      prev = NULL;
   }
}

void SxException::copy (const SxException &in)
{
   destroy ();
   init (in);
   SxException *dst = this;
   SxException *src = in.next;
   while (src)  {
      dst->next = new SxException();
      dst->next->init (*src);
      dst->next->prev = dst;
      dst = dst->next;
      src = src->next;
   }
   this->tail = dst;
}

void SxException::causeSegFault ()
{
//#  ifdef NDEBUG
//      std::cout ("Exceptions can be traced only in the SFHIngX DEBUG mode.\n");
//      dumpCoreFile = false;
//#  else
      std::cout <<  "Exceptions will cause segmentation faults and "
              "dump a core file.\n";
      dumpCoreFile = true;
//#  endif /* NDEBUG */
}


void SxException::throwExceptions ()
{
   dumpCoreFile = false;
}

void SxException::print (bool printSender) const
{
   const SxException *ptr = this;
   while (ptr)  {
      std::cerr << std::endl << ptr->message << std::endl;
      if (printSender)
         std::cerr << "This Exception was emitted from "
                   << ptr->senderFile
                   << " (line " << ptr->senderLine << ")\n";

      ptr = ptr->next;
   }
}

SxString SxException::toString (int mode,
                                const char delimiter_[]) const
{
   SX_CHECK (mode >= 0x00 && mode <= 0x03, mode);

   SxList<SxString> list;

   const SxException *ptr = this;
   while (ptr)  {
      SxString str, type;
      if (mode & DebugSimple)  {
         str = ptr->senderFile;
         ssize_t i = str.getSize () - 1;
         while (i >= 0 && str(i) != '\\' && str(i) != '/') i--;
         if (i >= 0) str = str.subString (i + 1);
         str = " (" + str + ":" + SxString(ptr->senderLine) + ")";
      }
      if (mode & Stack && ptr->tag)  {
         type = "[" + SxString(ptr->tag) + "] ";
      }
      str = type + ptr->message + str;
      list.prepend (str);
      ptr = ptr->next;
   }

   if (mode & Stack)  {
     int i=1;
     list.foreach ([&i](auto it) { *it = SxString(i++) + ": " + *it; });
     return SxString::join (list, delimiter_);
   } else {
     return list.last ();
   }
}

void SxException::printStack (const char delimiter_[]) const
{
   const SxException *ptr = this->tail;
   int counter = 1;
   int i = 0;
   int strSize = 0;
   while (ptr) {
      cerr << counter++ << ": ";
      if (ptr->tag)  cerr << "[" << ptr->tag << "] ";
      if (ptr->message) cerr << ptr->message << " ";
      strSize = static_cast<int>(strlen (ptr->senderFile));
      i = strSize - 1;
      while (i >= 0 && ptr->senderFile[i] != '\\' && ptr->senderFile[i] != '/') 
         i--;
      cerr << "(";
      while ((i+1) < strSize)  { cerr << ptr->senderFile[i+1]; i++; }
      cerr << ":" << ptr->senderLine << ")";
      cerr << delimiter_;
      ptr = ptr->prev;
   }
}

bool SxException::hasTag (const char *tag_) const
{
   SX_CHECK (tag_ != NULL);
   return (SxString (tag_) == tag);
}

bool SxException::findTag (const char *tag_) const
{
   SX_CHECK (tag_ != NULL);
   SxString needle = tag_;
   const SxException *ptr = this;
   while (ptr)  {
      if (ptr->getTag() == needle)  return true;
      ptr = ptr->next;
   }
   return false;
}

