// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_FUNCTOR_H_
#define _SX_FUNCTOR_H_

#include <string.h>
#include <SxError.h>
#include <SxConfig.h>
#include <iostream>


#   ifndef NDEBUG
       class SxFunctorDbgInfo
       {
          public:
             SxFunctorDbgInfo () {}
             SxFunctorDbgInfo (const char *tag_)
             {
                if (tag_ == NULL) {
                   tag[0] = '\0';
                } else {
                   strncpy (tag, tag_, 239);
                   tag[239] = '\0';
                }
             }
             ~SxFunctorDbgInfo () {}
             bool operator == (const SxFunctorDbgInfo &inf) {
                if ((strncmp(tag, inf.tag, 240) == 0))
                   return true;
                else
                   return false;
             }
             void setTag (const char *tag_) {
                if (tag_ == NULL) {
                   tag[0] = '\0';
                } else {
                   strncpy (tag, tag_, 239);
                   tag[239] = '\0';
                }
             }
             const char * getTag () const {
                return tag;
             }
          protected:
             char tag[240];
       };
#   endif /* NDEBUG */




/** \brief Empty data type to handle variadic template classes

    Use this class to define variatic template classes. For example
\code
template<class A, class B=SxNull,class C=SxNull>
class MyClass { };

template<class A>
class MyClass<A,SxNull,SxNull>
{
   public:
      someCode
};

template<class A, class B>
class MyClass<A,B,SxNull>
{
   public:
      someCode
};
\endcode
\author Sixten Boeck, boeck@mpie.de */
class SxNull { };

// --------------------------------------------------------------------------

/** \brief Basis class for all functors

    \b SxFunctorBase = S/PHI/nX Functor Base class

    This functor class is a container for bound function pointers. It stores
    the object (callee), the member function or the static global function.
    The information is anonymized by casting all types to void.

    \author Sixten Boeck, boeck@mpie.de */
class SxFunctorBase
{

#   ifndef NDEBUG
       protected:
          SxFunctorDbgInfo debugInf;
#   endif /* NDEBUG */

   public:

#      ifndef NDEBUG
       void setDebugInfo (const char *tag_) {
          debugInf.setTag (tag_);
       }
#      endif /* NDEBUG */
       friend std::ostream &operator<< (std::ostream &os,
                                        const SxFunctorBase &functor_);

      /** \brief Type of a pointer to member functions

          According to the C++ standard, all pointer-to-member-functions
          can be byte-converted via reinterpret_cast<> to
          pointer-to-member-functions of a different class and signature.
          Back-conversion yields the original pointer. We use this feature to
          store any pointer-to-member-function in a
          void (SxFunctorBase::*)() called "MemFunc". It can never be used
          as such, only after type-converting to the original class. This is
          done in SxMemTranslator::metaFunc .
          For more on the intricacies of the pointer-to-member-functions, see

          https://www.codeproject.com/Articles/7150/Member-Function-Pointers-and-the-Fastest-Possible

        */
      typedef void (SxFunctorBase::*MemFunc)();

      /** \brief Type of a pointer to static global functions
        */
      typedef void (*Func)();

      SxFunctorBase () : callee(0), func(0) { /* empty */ }

      /** \param func_    Pointer to global function
          */
      SxFunctorBase (Func *func_, const char *tag_ = NULL)
         : callee(0), 
           func(func_)
      {
#   ifndef NDEBUG
       setDebugInfo (tag_);
#   else
       SX_UNUSED(tag_);
#   endif /* NDEBUG */
      }

     /** \param callee_  Pointer to object or NULL for global functions
          \param func_    Pointer to member function
          \param size     size of function pointer */
      SxFunctorBase (const void *callee_, const MemFunc &func_,
                     const char *tag_ = NULL)
         : callee(const_cast<void *>(callee_)), 
           func(0)
      {
         SX_CHECK (callee_);
         memFunc = func_;
#   ifndef NDEBUG
         setDebugInfo (tag_);
#   else
         SX_UNUSED(tag_);
#   endif /* NDEBUG */
      }

      SxFunctorBase (const SxFunctorBase &in)
         : callee(in.callee)
      {
         if (callee)  {  // member function pointer
            memFunc = in.memFunc;
         } else {        // global function pointer
            func = in.func;
         }
#   ifndef NDEBUG
       setDebugInfo (in.debugInf.getTag());
#   endif /* NDEBUG */
      }

      void operator= (const SxFunctorBase &in)
      {
         callee = in.callee;
         if (callee)  {  // member function pointer
            memFunc = in.memFunc;
         } else {        // global function pointer
            func = in.func;
         }
#   ifndef NDEBUG
         setDebugInfo (in.debugInf.getTag());
#   endif /* NDEBUG */ 
      }

      /** \brief support 'if (callbackPtr)'
        */
      operator int() const
      {
         return (callee == NULL && func != NULL) || (callee != NULL);
      }

      bool operator== (const SxFunctorBase &in)
      {
         if (!callee)  {  // C function
            return (func == in.func && !in.callee);
         }  else  {  // member function
            return (callee == in.callee && memFunc == in.memFunc);
         }
      }

      bool operator!= (const SxFunctorBase &in)
      {
         return !operator== (in);
      }

      /** \brief Pointer to object or NULL for global functions
        */
      void *callee;

      /** \brief function or member function data
        */
      union {
         void *func;
         MemFunc memFunc;
      };
};

inline std::ostream &operator<< (std::ostream &os, const SxFunctorBase &functor_)
{
#  ifndef NDEBUG
      if ((functor_.debugInf.getTag ())[0] != '\0') {
         os << "SxFunctorBase : "
            << functor_.debugInf.getTag () << "\n";
      }
#  else
      SX_UNUSED (functor_);
      SX_EXIT;
#  endif
   return os;
}
// --------------------------------------------------------------------------

template<class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull, 
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxFunctor : public SxFunctorBase
{
   // empty
};

// --- SxFunctor<RES>
template<class RES>
class SxFunctor<RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() () const {
         return metaFunc (*this);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1>
template<class RES, class P1>
class SxFunctor<RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1) const {
            return metaFunc (*this, p1);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES (*MetaFunc)(const SxFunctorBase &, P1);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2>
template<class RES, class P1, class P2>
class SxFunctor<RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2) const {
         return metaFunc (*this, p1, p2);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES (*MetaFunc)(const SxFunctorBase &, P1, P2);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3>
template<class RES, class P1, class P2, class P3>
class SxFunctor<RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3) const {
            return metaFunc (*this, p1, p2, p3);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4>
template<class RES, class P1, class P2, class P3, class P4>
class SxFunctor<RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4) const {
        return metaFunc (*this, p1, p2, p3, p4); 
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3 , P4);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4,P5>
template<class RES, class P1, class P2, class P3, class P4, class P5>
class SxFunctor<RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5) const {
            return metaFunc (*this, p1, p2, p3, p4, p5);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3, P4, P5);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};

// --- SxFunctor<RES,P1,P2,P3,P4,P5,P6>
template<class RES, class P1, class P2, class P3, class P4, class P5, class P6>
class SxFunctor<RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctorBase
{
   public:
      SxFunctor () : SxFunctorBase () { /* empty */ }
      using SxFunctorBase::operator int;

      RES operator() (P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6) const {
            return metaFunc (*this, p1, p2, p3, p4, p5, p6);
      }

   protected:

      // meta function pointer (function pointer + functor basis)
      typedef RES ( *MetaFunc)(const SxFunctorBase &, P1, P2, P3, P4, P5, P6);

      SxFunctor (MetaFunc mf_, SxFunctorBase::Func *f_, const char *tag_=NULL)
         : SxFunctorBase (f_, tag_), metaFunc(mf_) { /* empty */ }

      SxFunctor (MetaFunc mf_, const void *c_, const SxFunctorBase::MemFunc &f_,
                 const char *tag_ = NULL)
         : SxFunctorBase (c_, f_, tag_), metaFunc(mf_) { /* empty */ }

   private:
      MetaFunc metaFunc;  // function pointer + functor basis
};
//----------------------------------------------------------------------------

// translate function pointers to functors

// SxFuncTranslator<Func, RES,...>
template<class Func, class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull, 
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxFuncTranslator : public SxFunctor<RES,P1,P2,P3,P4,P5,P6,Terminator>
{ /* empty */ };

// SxFuncTranslator<Func, RES>
template<class Func, class RES>
class SxFuncTranslator<Func,RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES> (metaFunc, 0, (const void *)func_, 0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)()>(functor_.func))();
#        else
            return (Func (functor_.func))();
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1>
template<class Func, class RES, class P1>
class SxFuncTranslator<Func,RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1>
{
   public:

      SxFuncTranslator (RES (*func_)(P1), const char *tag_ = NULL) 
         : SxFunctor<RES,P1> (metaFunc,
                              (SxFunctorBase::Func *)const_cast<void *>((const void *)func_), tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1)>(functor_.func))(p1);
#        else
            return (Func(functor_.func))(p1);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2>
template<class Func, class RES, class P1, class P2>
class SxFuncTranslator<Func,RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor< RES,P1,P2>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES,P1,P2> (metaFunc, 0, (const void *)func_, 0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1,P2)>(functor_.func))(p1,p2);
#        else
            return (Func (functor_.func))(p1, p2);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3>
template<class Func, class RES, class P1, class P2, class P3>
class SxFuncTranslator<Func,RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES,P1,P2,P3> (metaFunc, 0, (const void *)func_, 0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2, P3 p3)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1,P2,P3)>(functor_.func))(p1,p2,p3);
#        else
            return (Func (functor_.func))(p1, p2, p3);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4>
template<class Func, class RES, class P1, class P2, class P3, class P4>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES,P1,P2,P3,P4> (metaFunc, 0, (const void *)func_, 0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1,P2,P3,P4)>(functor_.func))(p1,p2,p3,p4);
#        else
            return (Func (functor_.func))(p1, p2, p3, p4);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5>
template<class Func, class RES, class P1, class P2, class P3, class P4, class P5>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES,P1,P2,P3,P4,P5> (metaFunc, 0, (const void *)func_,
                                          0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1,P2,P3,P4,P5)>(functor_.func))(p1,p2,p3,p4,p5);
#        else
            return (Func (functor_.func))(p1, p2, p3, p4, p5);
#        endif
      }
};

// SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,P6>
template<class Func, class RES, class P1, class P2, class P3, class P4, class P5, class P6>
class SxFuncTranslator<Func,RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5,P6>
{
   public:
      SxFuncTranslator (Func func_, const char *tag_ = NULL) 
         : SxFunctor<RES,P1,P2,P3,P4,P5,P6> (metaFunc, 0, (const void *)func_,
                                             0, tag_) 
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6)
      {
         SX_CHECK (functor_.func);
#        ifdef WIN32
            return (reinterpret_cast<RES (*)(P1,P2,P3,P4,P5,P6)>(functor_.func))(p1,p2,p3,p4,p5,p6);
#        else
            return (Func (functor_.func))(p1, p2, p3, p4, p5, p6);
#        endif
      }
};



// ---------------------------------------------------------------------------

/*
 SxIsCallable allows to check at compile time whether the type T has
 an public overloaded operator() function with specific signature.
 */
struct SxIsCallable {

   template <typename T, typename RES>
   class ZeroArgs {
      template <typename U, RES (U::*)()> class Check;
      template <typename U, RES (U::*)()const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1>
   class OneArgs {
      template <typename U, RES (U::*)(P1)> class Check;
      template <typename U, RES (U::*)(P1)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1, typename P2>
   class TwoArgs {
      template <typename U, RES (U::*)(P1, P2)> class Check;
      template <typename U, RES (U::*)(P1, P2)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1, typename P2, typename P3>
   class ThreeArgs {
      template <typename U, RES (U::*)(P1, P2, P3)> class Check;
      template <typename U, RES (U::*)(P1, P2, P3)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1, typename P2, typename P3,
             typename P4>
   class FourArgs {
      template <typename U, RES (U::*)(P1, P2, P3, P4)> class Check;
      template <typename U, RES (U::*)(P1, P2, P3, P4)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1, typename P2, typename P3,
             typename P4, typename P5>
   class FiveArgs {
      template <typename U, RES (U::*)(P1, P2, P3, P4, P5)> class Check;
      template <typename U, RES (U::*)(P1, P2, P3, P4, P5)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };

   template <typename T, typename RES, typename P1, typename P2, typename P3,
             typename P4, typename P5, typename P6>
   class SixArgs {
      template <typename U, RES (U::*)(P1, P2, P3, P4, P5, P6)> class Check;
      template <typename U, RES (U::*)(P1, P2, P3, P4, P5, P6)const> class ConstCheck;
      template <typename U> static char func(Check<U, &U::operator()> *);
      template <typename U> static char func(ConstCheck<U, &U::operator()> *);
      template <typename U> static int func(...);
      public:
         static const bool value = sizeof(func<T>(0)) == sizeof(char);
   };
};


/*
 SxCallWrapper allows compile time conditional call to an object's operator()
 function if the condition is set to true. If condition is false then function
 is not called.
 */
template<bool Condition>
class SxCallWrapper {
   public:
      template<typename T, typename RES = void, typename... Args>
      static void call (T obj, Args... args) {
         SX_UNUSED(obj);
         SX_UNUSED(args...);
         SX_EXIT;
      }
};

template<>
class SxCallWrapper<true> {
   public:
      template<typename T, typename RES = void, typename... Args>
      static RES call (T obj, Args... args) {
         return obj(args...);
      }
};

// ------------------------------------------------------------------------

template<class Callee, class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull,
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxLambdaTranslator : public SxFunctor<RES,P1,P2,P3,P4,P5,P6,Terminator>
{ /* empty */ };


template<class Callee, class RES>
class SxLambdaTranslator<Callee,RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES> (metaFunc, &callee_, 
                          SxFunctorBase::MemFunc (nullptr), 
                          tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool ZeroArgs = SxIsCallable::ZeroArgs<Callee, RES>::value;
         if (ZeroArgs) {
            return SxCallWrapper<ZeroArgs>::template
                   call<Callee, RES>(*callee);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1>
class SxLambdaTranslator<Callee,RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1> (metaFunc, &callee_, 
                              SxFunctorBase::MemFunc (nullptr), 
                              tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool OneArgs = SxIsCallable::OneArgs<Callee, RES,
                                                         P1>::value;
         if (OneArgs) {
            return SxCallWrapper<OneArgs>::template
                   call<Callee, RES, P1>(*callee, p1);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1, class P2>
class SxLambdaTranslator<Callee,RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1,P2> (metaFunc, &callee_, 
                                 SxFunctorBase::MemFunc (nullptr), 
                                 tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool TwoArgs = SxIsCallable::TwoArgs<Callee, RES,
                                                         P1, P2>::value;
         if (TwoArgs) {
            return SxCallWrapper<TwoArgs>::template
                   call<Callee, RES, P1, P2>(*callee, p1, p2);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1, class P2, class P3>
class SxLambdaTranslator<Callee,RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1,P2,P3> (metaFunc, &callee_, 
                                    SxFunctorBase::MemFunc (nullptr),
                                    tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool ThreeArgs = SxIsCallable::ThreeArgs<Callee, RES, P1, P2,
                                                    P3>::value;
         if (ThreeArgs) {
            return SxCallWrapper<ThreeArgs>::template
                   call<Callee, RES, P1, P2, P3>(*callee, p1, p2, p3);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1, class P2, class P3, class P4>
class SxLambdaTranslator<Callee,RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1,P2,P3,P4> (metaFunc, &callee_, 
                                       SxFunctorBase::MemFunc (nullptr),
                                       tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool FourArgs = SxIsCallable::FourArgs<Callee, RES, P1,
                                                           P2, P3, P4>::value;
         if (FourArgs) {
            return SxCallWrapper<FourArgs>::template
                   call<Callee, RES, P1, P2, P3, P4>(*callee, p1, p2,
                                                     p3, p4);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1, class P2, class P3, class P4, class P5>
class SxLambdaTranslator<Callee,RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1,P2,P3,P4,P5> (metaFunc, &callee_,
                                          SxFunctorBase::MemFunc(nullptr), tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool FiveArgs = SxIsCallable::FiveArgs<Callee, RES, P1, P2,
                                                           P3, P4, P5>::value;
         if (FiveArgs) {
            return SxCallWrapper<FiveArgs>::template
                   call<Callee, RES, P1, P2, P3, P4, P5>(*callee,
                                                         p1, p2, p3,
                                                         p4, p5);
         } else { SX_EXIT; }
      }
};

template<class Callee, class RES, 
         class P1, class P2, class P3, class P4, class P5, class P6>
class SxLambdaTranslator<Callee,RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5,P6>
{
   public:
      SxLambdaTranslator (Callee &callee_, const char *tag_ = NULL)
         : SxFunctor<RES,P1,P2,P3,P4,P5,P6> (metaFunc, &callee_,
                                             SxFunctorBase::MemFunc (nullptr),
                                             tag_)
      { }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6)
      {
         Callee *callee = (Callee *)functor_.callee;
         static const bool SixArgs = SxIsCallable::SixArgs<Callee, RES, P1, P2,
                                                         P3, P4, P5, P6>::value;
         if (SixArgs) {
            return SxCallWrapper<SixArgs>::template
                   call<Callee, RES, P1, P2, P3, P4, P5, P6>(*callee,
                                                             p1, p2, p3,
                                                             p4, p5, p6);
         } else { SX_EXIT; }
      }
};

// ------------------------------------------------------------------------

// translate member function pointers to functors
// This implementation allows to slightly modify the signature on the fly
// The bound pointer's calling signature is RES func(P1, ... P6)
// The member function's calling signature is TRES A::func (TP1, ..., TP6)
// i.e., P1 must be convertible to TP1, P2 to TP2, ...
//       and TRES must be convertible to RES
// The conversion and possible compile time errors occur in metaFunc.
// SxCBoundPtr takes care of getting the number of parameters right,
// and puts the member function's signature into SxMemTranslator's
// MemberFunc template parameter

// SxMemTranslator<Callee, MemberFunc, RES, ...>
template<class Callee, class MemberFunc, class RES, 
         class P1 = SxNull, class P2 = SxNull, class P3 = SxNull,
         class P4 = SxNull, class P5 = SxNull, class P6 = SxNull,
         class Terminator = SxNull
        >
class SxMemTranslator : public SxFunctor<RES,P1,P2,P3,P4,P5,P6,Terminator>
{ /* empty */ };


// SxMemTranslator<Callee, MemberFunc, RES>
template<class Callee, class MemberFunc, class RES>
class SxMemTranslator<Callee,MemberFunc,RES,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)();
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1>
template<class Callee, class MemberFunc, class RES, 
         class P1>
class SxMemTranslator<Callee,MemberFunc,RES,P1,SxNull,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1);
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1, P2>
template<class Callee, class MemberFunc, class RES, 
         class P1, class P2>
class SxMemTranslator<Callee,MemberFunc,RES,P1,P2,SxNull,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1,P2> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, P1 p1, P2 p2)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1, p2);
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1, P2, P3>
template<class Callee, class MemberFunc, class RES, 
         class P1, class P2, class P3>
class SxMemTranslator<Callee,MemberFunc,RES,P1,P2,P3,SxNull,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1,P2,P3> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1, p2, p3);
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1, P2, P3, P4>
template<class Callee, class MemberFunc, class RES, 
         class P1, class P2, class P3, class P4>
class SxMemTranslator<Callee,MemberFunc,RES,P1,P2,P3,P4,SxNull,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1,P2,P3,P4> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1, p2, p3, p4);
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1, P2, P3, P4, P5>
template<class Callee, class MemberFunc, class RES, 
         class P1, class P2, class P3, class P4, class P5>
class SxMemTranslator<Callee,MemberFunc,RES,P1,P2,P3,P4,P5,SxNull,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1,P2,P3,P4,P5> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1, p2, p3, p4, p5);
      }
};

// SxMemTranslator<Callee, MemberFunc, RES, P1, P2, P3, P4, P5, P6>
template<class Callee, class MemberFunc, class RES, 
         class P1, class P2, class P3, class P4, class P5, class P6>
class SxMemTranslator<Callee,MemberFunc,RES,P1,P2,P3,P4,P5,P6,SxNull>
   : public SxFunctor<RES,P1,P2,P3,P4,P5,P6>
{
   public:
      SxMemTranslator (Callee &callee_, const MemberFunc &memFunc_, const char *tag_=NULL)
         : SxFunctor<RES,P1,P2,P3,P4,P5,P6> (metaFunc, &callee_,
                           reinterpret_cast<SxFunctorBase::MemFunc>(memFunc_),
                           tag_)
      { /* empty */ }

      static RES metaFunc (const SxFunctorBase &functor_, 
                           P1 p1, P2 p2, P3 p3, P4 p4, P5 p5, P6 p6)
      {
         Callee *callee = (Callee *)functor_.callee;
         MemberFunc memFunc = reinterpret_cast<MemberFunc>(functor_.memFunc);
         return (callee->*memFunc)(p1, p2, p3, p4, p5, p6);
      }
};

#endif /* _SX_FUNCTOR_H_ */
