// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#include <SxMarker.h>
#include <SxUniqueList.h>
#include <SxCalc.h>
#include <SxRegex.h>

SxMarker::SxMarker (const SxString &border_)
{
   // for consistency: should be ***marker*** or ###marker###
   SX_CHECK_VAR (border_.getSize() >= 1, border_.getSize());

   borderDef = (border_.getSize() == 1) ? border_ + border_ + border_
                                        : border_;
}


SxMarker::SxMarker (const SxMap<SxString,SxString> &markers_,
                    const SxString &border_)
   : markers(markers_)
{
   // for consistency: should be ***marker*** or ###marker###
   SX_CHECK_VAR (border_.getSize() >= 1, border_.getSize());

   borderDef = (border_.getSize() == 1) ? border_ + border_ + border_
                                        : border_;
}

SxMarker::~SxMarker ()
{
   // empty
}

SxString &SxMarker::operator() (const SxString &key)
{
   return markers(key);
}

SxString SxMarker::operator() (const SxString &key) const
{
   return markers(key);
}


SxString SxMarker::function (const SxString &funcId, const SxString &) const
{
   return "***LOOKUP-ERROR:" + funcId + "() undefined***";
}

SxString SxMarker::apply (const SxString &in, const SxString &border_) const
{
   SxString border = borderDef;
   if (border_ != "")  border = border_;

   SxMap<SxString,SxString> m = markers;
   long tmpIdx = 1;

   SxString res = in;
   SxString marker, value, expr, pattern, markerStr;
   SxList<SxString>::Iterator keyIt;
   SxList<double> numVals;
   bool err = false;
   // --- replace unresolved symbols with default values
   SxList<SxString> matches, terms;
   float markerVal, exprVal;
   bool found = true;
   while (found)  {
      found = false;

      // --- replace markers with values
      SxList<SxString> keys = m.getKeys();
      for (keyIt=keys.begin(); keyIt != keys.end(); ++keyIt)  {
         marker = *keyIt;
         value  = m(marker);
         res    = res.substitute (border + marker + border, value);
         err    = false;
         double d = value.toDouble (&err);
         if (err)  numVals << -1.;
         else      numVals << d;
      }

      // --- strip borders
      marker   = res.right(border).left(border);
      pattern  = border + marker + border;
      if (marker.getSize() == 0)  { found = false; break; }

      // --- do not replace error messages
      if (pattern.contains (border+"ERROR:"))        { found = false; break; }
      if (pattern.contains (border+"LOOKUP-ERROR:")) { found = false; break; }

      // --- ***=expr***
      if (marker(0) == '=' && marker.getSize() > 0)  {
         expr = marker.subString(1);
         // search variables $ABC
         bool undefinedVar = false;
         while (true)  {
            try { matches = SxRegex("\\$([A-Za-z_]+[0-9]*)","sP").match (expr); }
            catch (SxException e) { e.print(); SX_EXIT; }
            if (matches.getSize() == 0)  { break; }
            SX_CHECK_VAR (matches.getSize() == 3, matches.getSize());
            ssize_t idx = keys.findPos(matches(1));
            if (idx == -1)  {  // variable $ABC not found
               value += "?" + matches(0) = "!";
               undefinedVar = true;
               expr = expr.substitute("$"+matches(1), "?" + matches(1) + "!");
               marker = marker.substitute("$"+matches(1), "?" + matches(1) + "!");
            }  else  {
               expr = expr.substitute("$"+matches(1), "$" + SxString(idx));
            }
         }
         if (!undefinedVar)  {
            SxCalc calc (numVals);
            res = res.substitute(pattern, calc.eval (expr + ";"));
         } else {
            res = res.substitute(pattern, marker);
         }
         found = true;
         continue;
      }

      // --- ***function(marker)***
      try  {
         matches = SxRegex("^([A-Za-z0-9_]+)\\((.*)\\)", "sP").match (marker);
      }
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 4)  { 
         SxString funcId = matches(1);
         SxString val = function (funcId, matches(2));
         if (!val.contains (border+"LOOKUP-ERROR:"))  {
            SxString tmpKey = "_TMP_MARKER_" + SxString(tmpIdx++);
            m(tmpKey) = val;
            res = res.substitute (funcId+"("+matches(2)+")", tmpKey);
            found = true;
            continue;
         }  else  {
            res = res.substitute (pattern, val);
         }
      }

      
      // --- ***expr?if_true:if_false***
      try { matches = SxRegex("([^\\?]+)\\?([^:]*):(.*)", "sP").match (marker);
         if (matches.getSize() == 5)  { 
            expr  = matches(1);

            // expr<=>value
            terms = SxRegex("([^!<=>]+)([!<=>]+)([^!<=>]+)", "sP").match (expr);
            if (terms.getSize() == 5)  {
               if (!m.containsKey(terms(1))) markerStr = "";
               else                          markerStr = m(terms(1));

               SxList<SxString> tokens =
                  SxRegex("\"([^\"]*)\"", "sP").match (terms(3));
               if (tokens.getSize() == 3)  terms(3) = "'" + tokens(1) + "'";
               tokens = SxRegex("'([^']*)'", "sP").match (terms(3));
               if (tokens.getSize() == 3)  { // string comparison
                  terms(3) = tokens(1);

                  if (terms(2) == "==")  {
                     value = (markerStr == terms(3))
                           ? matches(2)
                           : matches(3);
                  } else if (terms(2) == "!=")  {
                     value = (markerStr != terms(3))
                           ? matches(2)
                           : matches(3);
                  }
               } else  {                     // numerical comparison
                  markerVal = markerStr.toFloat();
                  exprVal   = terms(3).toFloat();
                  if (terms(2) == "<")  
                     value = (markerVal < exprVal) ? matches(2) : matches(3);
                  else if (terms(2) == "<=")  
                     value = (markerVal <= exprVal) ? matches(2) : 
                                      matches(3);
                  else if (terms(2) == ">=")  
                     value = (markerVal >= exprVal) ? matches(2) : 
                                      matches(3);
                  else if (terms(2) == ">")  
                     value = (markerVal > exprVal) ? matches(2) : matches(3);
                  else if (terms(2) == "==")  
                     value = (fabs(markerVal - exprVal) < 1e-10) ? matches(2) : matches(3);
                  else 
                     SX_EXIT;
               }
               value = value.substitute ("[#]", markerStr);
               res   = res.substitute(pattern, value);
               found = true;
               continue;
            }  else  {
               // check only if marker exists: ***marker?if_true:if_false***
               value = (m.containsKey(expr)) 
                     ? matches(2) 
                     : matches(3);
               value = value.substitute ("[#]", markerStr);
               res   = res.substitute(pattern, value);
               found = true;
               continue;
            }
         } 
      } catch (SxException e) { e.print (); SX_EXIT; }

      // --- ***marker:pre:default:post***
      try { matches = SxRegex("(.+):(.*):(.*):(.*)", "sP").match (marker); } 
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 6)  { 
         if (m.containsKey(matches(1)))
            value = matches(2) 
                  + m(matches(1)) 
                  + matches(4);
         else
            value = matches(3);
         res = res.substitute(pattern, value);
         found = true;
         continue;
      }

      // --- ***marker:pre:default***
      try { matches = SxRegex("(.+):(.*):(.*)", "sP").match (marker); }
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 5)  { 
         if (m.containsKey(matches(1)))
            value = matches(2) + m(matches(1));
         else
            value = matches(3);
         res = res.substitute(pattern, value);
         found = true;
         continue;
      }


      // --- ***marker:default***
      try { matches = SxRegex("(.+):(.+)", "sP").match (marker); } 
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 4)  { 
         value = (m.containsKey(matches(1)))
               ? m(matches(1))
               : matches(2);
         res = res.substitute(pattern, value);
         found = true;
         continue;
      }

      // --- ***marker[id]***
      try { matches = SxRegex("(.+)\\[(.+)\\]", "sP").match (marker); } 
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 4)  { 
         if (!m.containsKey(matches(1)))  {
            res = res.substitute(pattern, "LOOKUP-ERROR:UNKNOWN_MARKER");
            found = true;
            continue;
         }
         value = m(matches(1));
         err = false;
         int i = matches(2).toInt (&err);
         SxList<SxString> tokens = value.tokenize (',');
         if (abs(i) >= tokens.getSize() )  {
            res = ""; //res.substitute(pattern, "LOOKUP-ERROR:BOUNDARY");
            found = true;
            continue;
         }
         ssize_t idx = (i >= 0) ? i : tokens.getSize() + i;
         res = res.substitute(pattern, tokens(idx));
         found = true;
         continue;
      }

   }

   return res;
}

SxList<SxString> SxMarker::getUnresolved (const SxString &tmpl, 
                                          const SxString &border_,
                                          ssize_t maxSize)
{
   SxUniqueList<SxString> res;

   SxString delim;
   if (border_ == "*")  {
      delim = "\\*";
   } else {
      SX_EXIT;  // delimiter not supported
   }
   SxString border = delim + delim + delim;
   //               \\*\\*\\*([^  \\*    ]+)(\\*\\*\\*)
   SxString expr   = border+"([A-Z1-9_]{1,"+maxSize+"})("+border+")";

   SxList<SxString> matches;
   bool found = true;
   SxString t = tmpl, id;
   while (found)  {
      // --- ***marker***
      try { matches = SxRegex(expr, "sP").match (t); } 
      catch (SxException e) { e.print (); SX_EXIT; }
      if (matches.getSize() == 4)  {
         SxString m = matches(1);
         t = matches.last();  // $'

         // ***expr?true:false*** || ***=expr*** || ***function(val)***
         if (   m.contains ("?") || m.contains ("=")
             || m.contains ("(")) 
         {
            continue;
         }

         if ( (id = m.left(":")) != "" )  {
            res << id;  // ***marker:value***
         } else if ( (id = m.left("[")) != "" )  {
            res << id;  // ***marker[id]***
         } else {
            res << m;   // ***marker***
         }
         
         continue;
      }
      found = false;
   }
   return res;
}


SxMarker::Iterator SxMarker::begin ()
{
   return markers.begin();
}

SxMarker::ConstIterator SxMarker::begin () const
{
   return markers.begin();
}

SxMarker::Iterator SxMarker::end ()
{
   return markers.end();
}

SxMarker::ConstIterator SxMarker::end () const
{
   return markers.end();
}

SxList<SxString> SxMarker::getKeys () const
{
   return markers.getKeys ();
}

SxList<SxString> SxMarker::getValues () const
{
   return markers.getValues ();
}

void SxMarker::append (const SxMarker &in)
{
   markers.append (in.markers);
}

bool SxMarker::containsKey (const SxString &key) const
{
   return markers.containsKey (key);
}

void SxMarker::print () const
{
   SxMap<SxString,SxString>::ConstIterator it;
   for (it = markers.begin(); it != markers.end(); ++it)  {
      cout << it.getKey () << " = " << it.getValue () << endl;
   }
}
