
// ---------------------------------------------------------------------------
//
//      The ab-initio based multiscale library
//
//                  S / P H I / n X
//
//      Copyright:  Max-Planck-Institute for Iron Research
//                  40237 Duesseldorf, Germany
//
//      Contact:    https://sxlib.mpie.de
//      Authors:    see sphinx/AUTHORS
//      License:    see sphinx/LICENSE
//
// ---------------------------------------------------------------------------

#ifndef _SX_SIGNALS_H_
#define _SX_SIGNALS_H_

#include <SxCBoundPtr.h>
#include <SxList.h>
#include <iostream>
#include <string.h>
// --- support QT-like doxygen documentation 
#ifdef BUILD_DOXYGEN
#   define SX_SIGNAL(name)  name()
#else
#   define slots    /* empty */
#   undef signals
#   define signals  public /* empty */
#   define SX_SIGNAL(name)  name
#endif

#   ifndef NDEBUG
       class SxSigDbgInfo
       {
          public:
             SxSigDbgInfo () {}
             SxSigDbgInfo (const char *slot_, const char *tag_,
                           const char *signalName)
             {
                if (tag_ == NULL) {
                   tag[0] = '\0';
                } else {
                   strncpy (tag, tag_, 239);
                   tag[239] = '\0';
                }
                if (slot_ == NULL) {
                   slot[0] = '\0';
                } else {
                   strncpy (slot, slot_, 79);
                   slot[79] = '\0';
                }
                setSlotTag (signalName);
             }
             ~SxSigDbgInfo () {}
             bool operator == (const SxSigDbgInfo &inf) {
                if ((strncmp(tag, inf.tag, 240) == 0) &&
                    (strncmp(slot,inf.slot, 80) == 0))
                   return true;
                else
                   return false;
             }
             void setSlotTag (const char *signalName) {
                if (slot[0] == '\0') {
                   slotTag[0] = '\0';
                } else {
                   strcpy (slotTag, "'");
                   strcat (slotTag, slot);
                   strcat (slotTag, "' is connected to:\n- '");
                   strcat (slotTag, signalName);
                   strcat (slotTag, "' at ");
                   strcat (slotTag, tag);
                   slotTag[247] = '\0';
                }
             }
             const char *getTag () const {
                return tag;
             }
             const char *getSlot () const {
                return slot;
             }
             const char *getSlotTag () const {
                if (slotTag[0] == '\0')  return NULL;
                else                     return slotTag;
             }
          protected:
             char tag[240];
             char slot[80];
             char slotTag[248];
       };
#   endif /* NDEBUG */

/** \brief Signal/Slot handling

    \b SxSignals = S/PHI/nX Signals

    This class provides the base for a Signal/Slot mechanism in S/PHI/nX. 
\code
#include <SxSigSlots.h>
class MySlider
{
   public:
      MySlider ()
      {
         sxconnect (this, valueChanged, this, MySlider::slotABC);
      }

   public slots:
 
      void slotABC (double val)
      {
         printf ("AAA::slotABC received: %g\n", val);
      }

   public signals:
      SxSignal<void,double> SIGNAL(valueChanged);
};

class MyIsosurface
{
   public:
      MyIsosurface () { }
 
      void generateIso (double threshold)
      {
         printf ("BBB::slotABC received: %g\n", threshold);
      }

};


int main ()
{
   MySlider slider;
   MyIsosurface iso;

   sxconnect (&slider, valueChanged, &iso, MyIsosurface::generateIso);

// slider.valueChanged.connect (&fooIso);


   sxconnect (&slider, valueChanged, NULL, fooIso);
   slider.valueChanged.send (0.5);
}
\endcode
    \author Sixten Boeck, boeck@mpie.de */
// ---------------------------------------------------------------------------
template<class P1=SxNull, class P2=SxNull, class P3=SxNull, class P4=SxNull, class P5=SxNull, class P6=SxNull>
class SxSignal
{
   public:
      typedef SxCBoundPtr<void,P1,P2,P3,P4,P5,P6>  TBoundPtr;

#   ifndef NDEBUG
       void setSigName (const char *sigName) {
          strncpy (signalName, sigName, 79);
          signalName[79] = '\0';
       }
#   endif /* NDEBUG */

      friend std::ostream &operator<< (std::ostream &os,
                                       const SxSignal<P1,P2,P3,P4,P5,P6> &sig_)
      {
#     ifndef NDEBUG
         os << "'" << sig_.signalName 
            << "' is connected to: " 
            << std::endl;
         for (auto itInf = sig_.infoList.begin ();
              itInf != sig_.infoList.end (); ++itInf) 
         {
            if (((*itInf).getSlot ())[0] !=  '\0') {
               os << "- '" << (*itInf).getSlot () << 
                  "' at " << (*itInf).getTag () << std::endl;
            }
         }
#     else
         SX_UNUSED (sig_);
         SX_EXIT;
#     endif
         return os;
      } 

      // --- global function slots
      template<class TP1>
      void connect (void *dst, void (*slot)(TP1),
                    const char *slot_ = NULL,
                    const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
       infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
       SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class TP1,class TP2>
      void connect (void *dst, void (*slot)(TP1,TP2), 
                    const char *slot_ = NULL,
                    const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class TP1,class TP2,class TP3>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class TP1,class TP2,class TP3,class TP4>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class TP1,class TP2,class TP3,class TP4,class TP5>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4,TP5),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class TP1,class TP2,class TP3,class TP4,class TP5,class TP6>
      void connect (void *dst, void (*slot)(TP1,TP2,TP3,TP4,TP5,TP6),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         SX_CHECK (!dst);
         TBoundPtr bPtr = TBoundPtr::create (slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }

      // --- member function slots
      template<class DST,class Class,class TP1>
      void connect (DST *dst, void (Class::* const &slot)(TP1),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         if (!receivers.contains (bPtr)) { 
            receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */ 
         }
      }
      template<class DST,class Class,class TP1,class TP2>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4), 
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,
               class TP5>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED (slot_, tag_);
#   endif
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,
               class TP5,class TP6>
      void connect (DST *dst,
                    void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6),
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }

      // --- const member function slots
      template<class DST,class Class,class TP1>
      void connect (DST *dst, void (Class::* const &slot)(TP1)const,
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2)const,
                    const char *slot_ = NULL,const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3)const,
                    const char *slot_ = NULL,const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void connect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4)const,
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4,
               class TP5>
      void connect (DST *dst, 
                    void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5)const, 
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         TBoundPtr bPtr = TBoundPtr::create (*dst, slot);
         SX_CHECK (!receivers.contains (bPtr));
         receivers << bPtr;
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }
      template<class DST,class Class,class TP1,class TP2,class TP3,
               class TP4,class TP5,class TP6>
      void connect (DST *dst, 
                    void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6)const,
                    const char *slot_ = NULL, const char *tag_ = NULL)
      {
         receivers << TBoundPtr::create (*dst, slot);
#   ifndef NDEBUG
         infoList << SxSigDbgInfo (slot_, tag_, signalName);
#   else
         SX_UNUSED(slot_, tag_);
#   endif /* NDEBUG */
      }

      void slotDeregister (void *p, const char *)
      {
#   ifdef NDEBUG
       for (typename SxList<TBoundPtr>::Iterator it = receivers.begin();
            it != receivers.end (); 
            ++it)
       {
          if (it->callee == p)  {
             receivers.removeElement (*it);
             break;
          }
       }
#   else
       auto itInf = infoList.begin ();
       for (typename SxList<TBoundPtr>::Iterator it = receivers.begin();
            it != receivers.end (); 
            (++it,++itInf))
       {
          if (it->callee == p)  {
             receivers.removeElement (*it);
             infoList.removeElement (*itInf);
             break;
          }
       }
#   endif /* NDEBUG */
      }

#   ifndef NDEBUG
       void removeDebugInfo (const TBoundPtr &bPtr)
       {
          auto itInf = infoList.begin ();
          for (typename SxList<TBoundPtr>::Iterator it = receivers.begin();
               it != receivers.end (); 
               (++it,++itInf))
          {
             if (*it == bPtr)  {
                infoList.removeElement ((*itInf));
                break;
             }
          }
       }
#   endif /* NDEBUG */
      void disconnect (const TBoundPtr &cb)  
      {
         SX_CHECK (receivers.contains (cb));
#   ifndef NDEBUG
       removeDebugInfo (cb);
#   endif /* NDEBUG */
         receivers.removeElement (cb);
      }

      template<class DST,class Class,class TP1>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2,class TP3>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,class TP1,class TP2,class TP3,class TP4>
      void disconnect (DST *dst, void (Class::* const &slot)(TP1,TP2,TP3,TP4))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,
               class TP1,class TP2,class TP3,
               class TP4,class TP5>
      void disconnect (DST *dst, 
                       void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      template<class DST,class Class,
               class TP1,class TP2,class TP3,
               class TP4,class TP5,class TP6>
      void disconnect (DST *dst, 
                       void (Class::* const &slot)(TP1,TP2,TP3,TP4,TP5,TP6))
      {
         disconnect (TBoundPtr::create (*dst, slot));
      }

      void send ()
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)((*itInf).getSlotTag ());
#   endif
      }

      template<class TP1>
      void send (const TP1 &p1)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, (*itInf).getSlotTag ());
#   endif
      }

      template<class TP1,class TP2>
      void send (const TP1 &p1, const TP2 &p2)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, p2, (*itInf).getSlotTag ());
#   endif
      }

      template<class TP1,class TP2,class TP3>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, p2, p3, (*itInf).getSlotTag ());
#   endif
      }

      template<class TP1,class TP2,class TP3, class TP4>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, const TP4 &p4)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, p2, p3, p4, (*itInf).getSlotTag ());
#   endif
      }

      template<class TP1,class TP2,class TP3, class TP4, class TP5>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, 
                 const TP4 &p4, const TP5 &p5)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4, p5, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, p2, p3, p4, p5, (*itInf).getSlotTag ());
#   endif
      }

      template<class TP1,class TP2,class TP3, class TP4, class TP5, class TP6>
      void send (const TP1 &p1, const TP2 &p2, const TP3 &p3, 
                 const TP4 &p4, const TP5 &p5, const TP6 &p6)
      {
#   ifdef NDEBUG
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); ++it)
            (*it)(p1, p2, p3, p4, p5, p6, NULL);
#   else
         auto itInf = infoList.begin ();
         typename SxList<TBoundPtr>::Iterator it;
         for (it = receivers.begin(); it != receivers.end(); (++it,++itInf))
            (*it)(p1, p2, p3, p4, p5, p6, (*itInf).getSlotTag ());
#   endif
      }




   protected:
      SxList<TBoundPtr> receivers;
#   ifndef NDEBUG
       char signalName[80] = {'\0'};
       SxList<SxSigDbgInfo> infoList;
#   endif /* NDEBUG */
};


#endif /* _SX_SIGNALS_H_ */
